"""
Created on 28 Nov 2023

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

examples:
1.23
3.2.27

https://en.wikipedia.org/wiki/Software_versioning
"""

import re

from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class SoftwareVersion(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        match = re.match(r'^(\d+)\.(\d+)(?:\.(\d+))?', jdict)

        if not match:
            raise ValueError(jdict)

        return cls([int(part) for part in match.groups() if part is not None])


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, parts):
        """
        Constructor
        """
        self.__parts = parts                                # list of int


    def __len__(self):
        return len(self.__parts)


    def __eq__(self, other):
        try:
            if len(self) != len(other):
                return False

            for i in range(len(self)):
                if self.__parts[i] != other.__parts[i]:
                    return False

            return True

        except (TypeError, AttributeError):
            return False


    def __gt__(self, other):
        if len(self) != len(other):
            raise ValueError(other)

        for i in range(len(self)):
            if self.__parts[i] > other.__parts[i]:
                return True

            if self.__parts[i] < other.__parts[i]:
                return False

        return False


    def __lt__(self, other):
        if len(self) != len(other):
            raise ValueError(other)

        for i in range(len(self)):
            if self.__parts[i] < other.__parts[i]:
                return True

            if self.__parts[i] > other.__parts[i]:
                return False


        return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        return '.'.join([str(part) for part in self.__parts])


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def parts(self):
        return self.__parts


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "SoftwareVersion:{%s}" % self.parts



"""
Created on 12 Jan 2024

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

document example:
[{"rec": "2024-01-10T11:55:45Z", "event": "checksum error"}]
"""

from scs_core.csv.csv_log import CSVLog, LogEntry


# --------------------------------------------------------------------------------------------------------------------

class PSUEventLog(CSVLog):
    """
    classdocs
    """

    __FILENAME = "psu_event_log.csv"
    __MAX_ENTRIES = 5000                                # ≈ 180 KB

    @classmethod
    def persistence_location(cls):
        return cls.log_dir(), cls.__FILENAME


    @classmethod
    def max_permitted_entries(cls):
        return cls.__MAX_ENTRIES


    @classmethod
    def construct_entry(cls, event):
        return LogEntry.construct(event)


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return cls([]) if skeleton else None

        entries = [LogEntry.construct_from_jdict(entry_jdict) for entry_jdict in jdict]

        return cls(entries)



"""
Created on 17 Jun 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import requests

from scs_core.aws.client.api_client import APIClient
from scs_core.aws.config.endpoint import APIEndpoint

from scs_core.aws.monitor.alert.alert_specification_intercourse import AlertSpecificationFindRequest, \
    AlertSpecificationFindResponse

from scs_core.data.json import JSONify


# --------------------------------------------------------------------------------------------------------------------

class Endpoint(APIEndpoint):
    @classmethod
    def configuration(cls):
        return cls('AlertSpecAPI/AlertSpecification',
                   'https://a066wbide8.execute-api.us-west-2.amazonaws.com/default/AlertSpecification')


# --------------------------------------------------------------------------------------------------------------------

class AlertSpecificationManager(APIClient):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self):
        super().__init__()


    # ----------------------------------------------------------------------------------------------------------------

    def find(self, token, description_filter, topic_filter, path_filter, creator_filter):
        request = AlertSpecificationFindRequest(description_filter, topic_filter, path_filter, creator_filter)

        response = requests.get(Endpoint.url(), headers=self._token_headers(token), params=request.params())
        self._check_response(response)

        return AlertSpecificationFindResponse.construct_from_jdict(response.json())


    def retrieve(self, token, id):
        url = Endpoint.url(id)

        http_response = requests.get(url, headers=self._token_headers(token))
        self._check_response(http_response)

        response = AlertSpecificationFindResponse.construct_from_jdict(http_response.json())

        return response.alerts[0] if response.alerts else None


    def create(self, token, alert):
        http_response = requests.post(Endpoint.url(), headers=self._token_headers(token), data=JSONify.dumps(alert))
        self._check_response(http_response)

        response = AlertSpecificationFindResponse.construct_from_jdict(http_response.json())

        return response.alerts[0] if response.alerts else None


    def update(self, token, alert):
        url = Endpoint.url(alert.id)

        http_response = requests.post(url, headers=self._token_headers(token), data=JSONify.dumps(alert))
        self._check_response(http_response)

        response = AlertSpecificationFindResponse.construct_from_jdict(http_response.json())

        return response.alerts[0] if response.alerts else None


    def delete(self, token, id):
        url = Endpoint.url(id)

        http_response = requests.delete(url, headers=self._token_headers(token))
        self._check_response(http_response)



"""
Created on 5 Mar 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://www.airnow.gov/
"""

import re

from scs_core.aqcsv.specification.country_numeric import CountryNumeric

from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class AQCSVSite(JSONable):
    """
    classdocs
    """

    LOCATION_CODE_LENGTH = 9

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_code(cls, code):
        if code is None:
            return None
        try:
            match = re.match(r'(\d{3})(MM)?(\d{9}).*', str(code))
        except TypeError:
            raise ValueError(code)

        if match is None:
            raise ValueError(code)

        fields = match.groups()

        country_code = fields[0]
        location_code = fields[2]
        is_mobile = fields[1] is not None

        return AQCSVSite(country_code, location_code, is_mobile)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, country_code, location_code, is_mobile):
        """
        Constructor
        """
        self.__country_code = int(country_code)             # int(3)
        self.__location_code = location_code                # string
        self.__is_mobile = is_mobile                        # bool


    def __eq__(self, other):
        try:
            return \
                self.country_code == other.country_code and \
                self.location_code == other.location_code and \
                self.is_mobile == other.is_mobile

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        return self.as_code()


    def as_code(self):
        mobility_code = 'MM' if self.is_mobile else ''

        return str(self.country_code) + str(mobility_code) + str(self.location_code)


    # ----------------------------------------------------------------------------------------------------------------

    def country(self):
        return CountryNumeric.instance(self.country_code)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def country_code(self):
        return self.__country_code


    @property
    def location_code(self):
        return self.__location_code


    @property
    def is_mobile(self):
        return self.__is_mobile


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "AQCSVSite:{country_code:%03d, location_code:%s, is_mobile:%s}" % \
               (self.country_code, self.location_code, self.is_mobile)



"""
Created on 20 Jan 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

The Tail class provides similar functionality to the Unix tail CLU, but with significant differences.

The Tail readlines() method yields successive lines as they are added to the file being tailed. When the file
is closed by the writer process, and following the next read by the client, the readlines() method exits
with a RuntimeError (caused by a StopIteration).

Important note: readlines() ends irrespective of whether the client has read all the lines in the file!

https://stackoverflow.com/questions/16175745/read-new-line-with-pynotify
https://www.linode.com/docs/development/monitor-filesystem-events-with-pyinotify/
"""

import sys

from pyinotify import EventsCodes, Notifier, ProcessEvent, WatchManager


# --------------------------------------------------------------------------------------------------------------------

class Tail(object):
    """
    classdocs
    """

    DEFAULT_TIMEOUT =   600000                  # 10 minutes in milliseconds

    # ----------------------------------------------------------------------------------------------------------------

    __IN_MODIFY =       EventsCodes.FLAG_COLLECTIONS['OP_FLAGS']['IN_MODIFY']
    __IN_CLOSE_WRITE =  EventsCodes.FLAG_COLLECTIONS['OP_FLAGS']['IN_CLOSE_WRITE']
    __IN_MOVED_TO =     EventsCodes.FLAG_COLLECTIONS['OP_FLAGS']['IN_MOVED_TO']

    __IN_EVENTS =       __IN_MODIFY | __IN_CLOSE_WRITE | __IN_MOVED_TO


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, path, timeout=None):
        handler = TailEventHandler(path)

        watch_manager = WatchManager()
        watch_manager.add_watch(path, cls.__IN_EVENTS)

        notifier_timeout = cls.DEFAULT_TIMEOUT if timeout is None else timeout
        notifier = TailNotifier(watch_manager, default_proc_fun=handler, timeout=notifier_timeout)

        return cls(notifier, handler)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, notifier, handler):
        """
        Constructor
        """
        self.__notifier = notifier                              # TailNotifier
        self.__handler = handler                                # TailEventHandler


    def __iter__(self):
        return self.readlines()


    # ----------------------------------------------------------------------------------------------------------------

    def open(self):
        self.__handler.open_file()


    def close(self):
        self.__handler.close_file()


    def readlines(self):
        # existing lines...
        for line in self.__handler.read_head():
            yield line

        # new lines...
        for line in self.__notifier.loop(self.__handler.read_tail):
            yield line

        # raise StopIteration


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Tail:{handler:%s}" %  self.__handler


# --------------------------------------------------------------------------------------------------------------------

class TailEventHandler(ProcessEvent):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, path):
        """
        Constructor
        """
        super().__init__()

        self.__path = path

        self.__file = None
        self.__terminate = False


    # ----------------------------------------------------------------------------------------------------------------

    def open_file(self):
        self.__file = open(self.__path)


    def close_file(self):
        if self.__file is None:
            return

        self.__file.close()
        self.__file = None


    def read_head(self, _notifier=None):
        for line in self.__file.readlines():
            yield line.strip()


    def read_tail(self, _notifier=None):
        if self.__terminate:
            return True

        return tuple(line.strip() for line in self.__file.readlines())


    # ----------------------------------------------------------------------------------------------------------------

    # noinspection PyPep8Naming
    def process_IN_MODIFY(self, _event):
        # self.__report("IN_MODIFY")
        pass


    # noinspection PyPep8Naming
    def process_IN_CLOSE_WRITE(self, _event):
        self.__report("IN_CLOSE_WRITE")

        self.__terminate = True


    # noinspection PyPep8Naming
    def process_IN_MOVED_TO(self, _event):
        self.__report("IN_MOVED_TO")

        self.__terminate = True


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def __report(cls, event_name):
        print("TailEventHandler: %s" % event_name, file=sys.stderr)
        sys.stderr.flush()


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "TailEventHandler:{path:%s, terminate:%s}" % (self.__path, self.__terminate)


# --------------------------------------------------------------------------------------------------------------------

class TailNotifier(Notifier):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, watch_manager, default_proc_fun=None, read_freq=0, threshold=0, timeout=None):
        super().__init__(watch_manager, default_proc_fun, read_freq, threshold, timeout)


    # ----------------------------------------------------------------------------------------------------------------

    def loop(self, callback=None, _daemonize=False, **_args):
        try:
            while True:
                self.process_events()

                response = callback(self)

                if response is True:
                    break                                   # file closed

                for line in response:
                    yield line

                if not self.check_events():
                    raise TimeoutError                      # timeout

                self.read_events()

        except (ConnectionError, KeyboardInterrupt, SystemExit):
            pass

        finally:
            self.stop()



"""
Created on 17 Jun 2023

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import requests

from scs_core.aws.client.api_client import APIClient
from scs_core.aws.config.endpoint import APIEndpoint

from scs_core.aws.monitor.device.device_monitor_specification import DeviceMonitorSpecification, \
    DeviceMonitorSpecificationList

from scs_core.data.json import JSONify

from scs_core.email.email import EmailRecipient


# --------------------------------------------------------------------------------------------------------------------

class Endpoint(APIEndpoint):
    @classmethod
    def configuration(cls):
        return cls('DevMonSpecAPI/DeviceMonitorSpecification',
                   'https://psnunpg4gb.execute-api.us-west-2.amazonaws.com/default/DeviceMonitorSpecification')


# --------------------------------------------------------------------------------------------------------------------

class DeviceMonitorSpecificationManager(APIClient):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self):
        super().__init__()


    # ----------------------------------------------------------------------------------------------------------------

    def find(self, token, email_address=None, device_tag=None, exact=False):
        payload = {
            'email_address': email_address,
            'device_tag': device_tag,
            'exact': exact
        }

        response = requests.get(Endpoint.url(), headers=self._token_headers(token), json=payload)
        self._check_response(response)

        return DeviceMonitorSpecificationList.construct_from_jdict(response.json())


    def add(self, token, device_tag, recipient: EmailRecipient):
        payload = {
            'device_tag': device_tag,
            'recipient': recipient
        }

        response = requests.post(Endpoint.url(), headers=self._token_headers(token), data=JSONify.dumps(payload))
        self._check_response(response)

        return DeviceMonitorSpecification.construct_from_jdict(response.json())


    def set_suspended(self, token, device_tag, is_suspended):
        payload = {
            'device_tag': device_tag,
            'is_suspended': is_suspended
        }

        response = requests.patch(Endpoint.url(), headers=self._token_headers(token), json=payload)
        self._check_response(response)

        return DeviceMonitorSpecification.construct_from_jdict(response.json())


    def remove(self, token, email_address, device_tag=None):
        payload = {
            'email_address': email_address,
            'device_tag': device_tag
        }

        response = requests.delete(Endpoint.url(), headers=self._token_headers(token), json=payload)
        self._check_response(response)

        return DeviceMonitorSpecificationList.construct_from_jdict(response.json())



"""
Created on 2 Oct 2023

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

from collections import OrderedDict
from http import HTTPStatus

from scs_core.aws.data.http_response import HTTPResponse
from scs_core.aws.monitor.alert.alert import AlertSpecification

from scs_core.client.http_exception import HTTPException

from scs_core.data.str import Str


# --------------------------------------------------------------------------------------------------------------------

class AlertSpecificationFindRequest(object):
    """
    classdocs
    """

    DESCRIPTION_FILTER = 'description'
    TOPIC_FILTER = 'topic'
    PATH_FILTER = 'path'
    CREATOR_FILTER = 'creator'

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_qsp(cls, qsp):
        if not qsp:
            return None

        description_filter = qsp.get(cls.DESCRIPTION_FILTER)
        topic_filter = qsp.get(cls.TOPIC_FILTER)
        path_filter = qsp.get(cls.PATH_FILTER)
        creator_filter = qsp.get(cls.CREATOR_FILTER)

        return cls(description_filter, topic_filter, path_filter, creator_filter)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, description_filter, topic_filter, path_filter, creator_filter):
        """
        Constructor
        """
        self.__description_filter = description_filter              # string
        self.__topic_filter = topic_filter                          # string
        self.__path_filter = path_filter                            # string
        self.__creator_filter = creator_filter                      # string


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def is_valid(cls):
        return True


    # ----------------------------------------------------------------------------------------------------------------

    def params(self):
        description_filter = self.description_filter if self.description_filter else None

        params = {
            self.DESCRIPTION_FILTER: description_filter,
            self.TOPIC_FILTER: self.topic_filter,
            self.PATH_FILTER: self.path_filter,
            self.CREATOR_FILTER: self.creator_filter
        }

        return params


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def description_filter(self):
        return self.__description_filter


    @property
    def topic_filter(self):
        return self.__topic_filter


    @property
    def path_filter(self):
        return self.__path_filter


    @property
    def creator_filter(self):
        return self.__creator_filter


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "AlertSpecificationFindRequest:{description_filter:%s, topic_filter:%s, path_filter:%s, " \
               "creator_filter:%s}" % \
               (self.description_filter, self.topic_filter, self.path_filter,
                self.creator_filter)


# --------------------------------------------------------------------------------------------------------------------

class AlertSpecificationFindResponse(HTTPResponse):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        status = HTTPStatus(jdict.get('statusCode'))

        if status != HTTPStatus.OK:
            raise HTTPException.construct(status.value, status.phrase, status.description)

        alerts = []
        if jdict.get('Items'):
            for alert_jdict in jdict.get('Items'):
                alerts.append(AlertSpecification.construct_from_jdict(alert_jdict))

        next_url = jdict.get('next')

        return cls(status, alerts, next_url=next_url)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, status, alerts, next_url=None):
        """
        Constructor
        """
        super().__init__(status)

        self.__alerts = alerts                  # array of Alert
        self.__next_url = next_url              # URL string


    def __len__(self):
        return len(self.alerts)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['statusCode'] = self.status.value

        if self.alerts is not None:
            jdict['Items'] = self.alerts

        if self.next_url is not None:
            jdict['next'] = self.next_url

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def alerts(self):
        return self.__alerts


    @property
    def next_url(self):
        return self.__next_url


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "AlertSpecificationFindResponse:{status:%s, alerts:%s, next_url:%s}" % \
               (self.status, Str.collection(self.alerts), self.next_url)



"""
Created on 19 Mar 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://stackoverflow.com/questions/7266558/pyserial-buffer-wont-flush
"""

import time

from abc import ABC, abstractmethod


# --------------------------------------------------------------------------------------------------------------------

class Serial(ABC):
    """
    classdocs
    """

    _DEFAULT_EOL = "\n"

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, device_identifier, baud_rate, hard_handshake=False):
        """
        Constructor
        """
        self._device_identifier = device_identifier             # string device path or int port number
        self._baud_rate = baud_rate                             # int
        self._hard_handshake = hard_handshake                   # bool

        self._ser = None


    # ----------------------------------------------------------------------------------------------------------------

    @abstractmethod
    def open(self, lock_timeout, comms_timeout):
        pass


    @abstractmethod
    def close(self):
        pass


    # ----------------------------------------------------------------------------------------------------------------

    def read_lines(self, eol=None, timeout=None):
        while True:
            try:
                yield self.read_line(eol, timeout)

            except TimeoutError:
                return


    def read_line(self, eol=None, timeout=None):
        terminator = self._DEFAULT_EOL if eol is None else eol
        end_time = None if timeout is None else time.time() + timeout

        line = ""
        while True:
            if timeout is not None and time.time() > end_time:
                raise TimeoutError(timeout)

            char = self._ser.read().decode(errors='ignore')
            line += char

            if line.endswith(terminator):
                break

        return line.strip()


    def write_line(self, text, eol=None):
        terminator = self._DEFAULT_EOL if eol is None else eol

        text_ln = text.strip() + terminator
        packet = text_ln.encode()

        return self._ser.write(packet)


    # ----------------------------------------------------------------------------------------------------------------

    def read(self, count):
        chars = self._ser.read(count)

        return chars


    def write(self, *chars):
        self._ser.write(bytearray(chars))


    def flush_input(self):
        self._ser.flushInput()


    def flush_output(self):
        self._ser.flushOutput()


    # ----------------------------------------------------------------------------------------------------------------

    @property
    @abstractmethod
    def device_identifier(self):
        return None


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        is_open = False if self._ser is None else self._ser.is_open

        return self.__class__.__name__ + ":{device_identifier:%s, baud_rate=%d, hard_handshake=%s, serial_open:%s}" % \
            (self._device_identifier, self._baud_rate, self._hard_handshake, is_open)



"""
Created on 21 Jun 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example JSON:
{"mode": "SYS", "device-name": "SCS Praxis/Handheld (dev)", "startup-message": "ON", "shutdown-message": "STANDBY",
"show-time": true}
"""

from collections import OrderedDict

from scs_core.data.json import PersistentJSONable


# --------------------------------------------------------------------------------------------------------------------

class DisplayConf(PersistentJSONable):
    """
    classdocs
    """

    __MODES = ['SYS']

    @classmethod
    def modes(cls):
        return cls.__MODES


    __FILENAME = "system_display_conf.json"

    @classmethod
    def persistence_location(cls):
        return cls.conf_dir(), cls.__FILENAME


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        mode = jdict.get('mode')

        device_name = jdict.get('device-name')
        startup_message = jdict.get('startup-message')
        shutdown_message = jdict.get('shutdown-message')

        show_time = jdict.get('show-time', True)

        return cls(mode, device_name, startup_message, shutdown_message, show_time)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, mode, device_name, startup_message, shutdown_message, show_time):
        """
        Constructor
        """
        super().__init__()

        self.__mode = mode                                          # string

        self.__device_name = device_name                            # string
        self.__startup_message = startup_message                    # string
        self.__shutdown_message = shutdown_message                  # string

        self.__show_time = show_time                                # bool


    def __eq__(self, other):
        try:
            return self.mode == other.mode and \
                   self.device_name == other.device_name and \
                   self.startup_message == other.startup_message and \
                   self.shutdown_message == other.shutdown_message and \
                   self.show_time == other.show_time

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def mode(self):
        return self.__mode


    @property
    def device_name(self):
        return self.__device_name


    @property
    def startup_message(self):
        return self.__startup_message


    @property
    def shutdown_message(self):
        return self.__shutdown_message


    @property
    def show_time(self):
        return self.__show_time


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['mode'] = self.mode

        jdict['device-name'] = self.device_name
        jdict['startup-message'] = self.startup_message
        jdict['shutdown-message'] = self.shutdown_message

        jdict['show-time'] = self.show_time

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "DisplayConf(core):{mode:%s, device_name:%s, startup_message:%s, shutdown_message:%s, " \
               "show_time:%s}" %  \
               (self.mode, self.device_name, self.startup_message, self.shutdown_message,
                self.show_time)



"""
Created on 2 Dec 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example request:
{"particulates":
    {"tag": "scs-be2-3", "src": "N3", "rec": "2020-08-16T07:52:24Z",
    "val": {"per": 4.9, "pm1": 17.8, "pm2p5": 19.4, "pm10": 19.5,
    "bin": [703, 32, 3, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    "mtf1": 79, "mtf3": 80, "mtf5": 0, "mtf7": 0, "sfr": 0.52,
    "sht": {"hmd": 45.8, "tmp": 30.1}},
    "exg": {"ISLin/N3/vPLHR": {"pm1": 41.0, "pm2p5": 21.0, "pm10": 15.3}}},
"climate":
    {"hmd": 60.5, "tmp": 25.9}}
"""

import json

from scs_core.comms.uds_client import UDSClient

from scs_core.data.json import JSONify

from scs_core.model.pmx.pmx_inference_client import PMxInferenceClient
from scs_core.model.pmx.s1.pmx_request import PMxRequest

from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class S1PMxInferenceClient(PMxInferenceClient):
    """
    classdocs
    """

    @classmethod
    def construct(cls, socket, inference_uds_path):
        # UDS...
        uds_client = UDSClient(socket, inference_uds_path)

        return cls(uds_client)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, uds_client):
        """
        Constructor
        """
        super().__init__(uds_client)

        self.__logger = Logging.getLogger()


    # ----------------------------------------------------------------------------------------------------------------

    def infer(self, opc_sample, ext_sht_datum):
        # self.__logger.error("opc_sample: %s" % opc_sample)
        # self.__logger.error("ext_sht_datum: %s" % ext_sht_datum)

        # request...
        pmx_request = PMxRequest(opc_sample, ext_sht_datum)
        # self.__logger.info("pmx_request: %s" % pmx_request)

        self._uds_client.request(JSONify.dumps(pmx_request.as_json()))
        response = self._uds_client.wait_for_response()

        return json.loads(response)


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "S1PMxInferenceClient:{uds_client:%s}" %  self._uds_client



"""
Created on 21 Sep 2020

@author: Jade Page (jade.page@southcoastscience.com)

DESCRIPTION User defined exceptions for the aws_group_setup script
"""


# --------------------------------------------------------------------------------------------------------------------

class Error(Exception):
    """Base class for other exceptions"""
    pass


class ProjectMissingError(Error):
    """Raised when the project read fails"""
    pass



"""
Created on 5 Apr 2022

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import requests

from scs_core.aws.client.api_client import APIClient
from scs_core.aws.config.endpoint import APIEndpoint
from scs_core.aws.security.cognito_device import CognitoDeviceIdentity

from scs_core.data.json import JSONify


# --------------------------------------------------------------------------------------------------------------------

class Endpoint(APIEndpoint):
    @classmethod
    def configuration(cls):
        return cls('CogDevAPI/CognitoDevices',
                   'https://6c2sfqt656.execute-api.us-west-2.amazonaws.com/default/CognitoDevices')


# --------------------------------------------------------------------------------------------------------------------

class CognitoDeviceManager(APIClient):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self):
        super().__init__()


    # ----------------------------------------------------------------------------------------------------------------

    def update_self(self, token, credentials):
        url = Endpoint.url('self')

        response = requests.patch(url, headers=self._token_headers(token), data=JSONify.dumps(credentials))
        self._check_response(response)

        return CognitoDeviceIdentity.construct_from_jdict(response.json())


    # ----------------------------------------------------------------------------------------------------------------

    def create(self, token, identity):
        response = requests.post(Endpoint.url(), headers=self._token_headers(token), data=JSONify.dumps(identity))
        self._check_response(response)

        return CognitoDeviceIdentity.construct_from_jdict(response.json())


    def update(self, token, identity):
        response = requests.patch(Endpoint.url(), headers=self._token_headers(token), data=JSONify.dumps(identity))
        self._check_response(response)

        return CognitoDeviceIdentity.construct_from_jdict(response.json())


    def delete(self, token, device_tag):
        payload = {"DeviceTag": device_tag}

        response = requests.delete(Endpoint.url(), headers=self._token_headers(token), data=JSONify.dumps(payload))
        self._check_response(response)

        # TODO: delete device from organisations?



"""
Created on 14 Oct 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

JSON example:
{"filesystem": "/dev/mmcblk0p1", "total": 15384184, "used": 319296, "free": 14892092,
"mounted-on": "/srv/SCS_logging", "is-available": false}

https://stackoverflow.com/questions/35469685/in-python-how-do-i-get-a-list-of-all-partitions-in-mac-os-x
"""

import os
import re

from collections import OrderedDict

from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class DiskVolume(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_df_row(cls, row):
        match = re.search(r'([^ ]+)\s+([0-9]+)\s+([0-9]+)\s+([0-9]+)\s+([0-9]+)%\s+([^ ]+)', row)

        if match is None:
            return None

        groups = match.groups()

        filesystem = groups[0]
        total = groups[1]
        used = groups[2]
        free = groups[3]
        mounted_on = groups[5]

        return cls(filesystem, free, used, total, mounted_on)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, filesystem, free, used, total, mounted_on):
        """
        Constructor
        """
        self.__filesystem = filesystem                      # string
        self.__total = int(total)                           # int blocks
        self.__used = int(used)                             # int blocks
        self.__free = int(free)                             # int blocks
        self.__mounted_on = mounted_on                      # string


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['filesystem'] = self.filesystem
        jdict['free'] = self.free
        jdict['used'] = self.used
        jdict['total'] = self.total
        jdict['mounted-on'] = self.mounted_on

        jdict['is-available'] = self.is_available

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def percent_used(self):
        if not self.is_available:
            return None

        percent = self.used / self.total * 100

        return round(percent, 1)


    @property
    def is_available(self):
        try:
            os.listdir(self.mounted_on)
            return True

        except OSError:
            return False


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def filesystem(self):
        return self.__filesystem


    @property
    def free(self):
        return self.__free


    @property
    def used(self):
        return self.__used


    @property
    def total(self):
        return self.__total


    @property
    def mounted_on(self):
        return self.__mounted_on


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "DiskVolume:{filesystem:%s, free:%s, used:%s, total:%s, mounted_on:%s}" %  \
               (self.filesystem, self.free, self.used, self.total, self.mounted_on)


# --------------------------------------------------------------------------------------------------------------------

class ReportedDiskVolume(DiskVolume):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        filesystem = jdict.get('filesystem')
        free = jdict.get('free')
        used = jdict.get('used')
        total = jdict.get('total')
        mounted_on = jdict.get('mounted-on')

        is_available = jdict.get('is-available')

        return cls(filesystem, free, used, total, mounted_on, is_available)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, filesystem, free, used, total, mounted_on, is_available):
        """
        Constructor
        """
        super().__init__(filesystem, free, used, total, mounted_on)

        self.__is_available = is_available                  # bool


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def is_available(self):
        return self.__is_available


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ReportedDiskVolume:{filesystem:%s, free:%s, used:%s, total:%s, mounted_on:%s, is_available:%s}" %  \
               (self.filesystem, self.free, self.used, self.total, self.mounted_on, self.is_available)



"""
Created on 17 Feb 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

Note: The system_serial_number is independent of both dfe serial number and host serial number. It is typically
associated with the hostname. The combination vendor-id + model-id + system-sn must be universally unique.

examples:
{"vendor-id": "scs", "model-id": "ap1", "model": "Alpha Pi Eng", "config": "V1", "system-sn": 6}
{"vendor-id": "SCS", "model-id": "BGB", "model": "Praxis", "config": "BGB", "system-sn": 406}
"""

from collections import OrderedDict

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import PersistentJSONable


# --------------------------------------------------------------------------------------------------------------------

class SystemID(PersistentJSONable):
    """
    classdocs
    """

    __FILENAME = "system_id.json"

    @classmethod
    def persistence_location(cls):
        return cls.conf_dir(), cls.__FILENAME


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        last_modified = LocalizedDatetime.construct_from_jdict(jdict.get('set-on'))

        vendor_id = jdict.get('vendor-id')
        model_id = jdict.get('model-id')

        model_name = jdict.get('model')
        configuration = jdict.get('config')

        system_serial_number = jdict.get('system-sn')

        return SystemID(vendor_id, model_id, model_name, configuration, system_serial_number,
                        last_modified=last_modified)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, vendor_id, model_id, model_name, configuration, system_serial_number,
                 last_modified=None):
        """
        Constructor
        """
        super().__init__(last_modified=last_modified)

        self.__vendor_id = vendor_id                                # string (3 chars)
        self.__model_id = model_id                                  # string (3 chars)

        self.__model_name = model_name                              # string
        self.__configuration = configuration                        # string

        self.__system_serial_number = system_serial_number          # string (by convention, int)


    def __eq__(self, other):
        try:
            return bool(self.last_modified) == bool(other.last_modified) and \
                   self.vendor_id == other.vendor_id and self.model_id == other.model_id and \
                   self.model_name == other.model_name and self.configuration == other.configuration and \
                   self.system_serial_number == other.system_serial_number

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def type_label(self):
        if self.model_name is None or self.configuration is None:
            return None

        return self.model_name + '/' + self.configuration


    def box_label(self):
        if self.model_name is None or self.configuration is None or self.system_serial_number is None:
            return None

        box_system_serial_number = str(self.system_serial_number).rjust(6, '0')

        return self.model_name + '/' + self.configuration + ' ' + box_system_serial_number


    def topic_label(self):
        if self.model_name is None or self.system_serial_number is None:
            return None

        topic_model_name = self.model_name.replace('/', '-').replace(' ', '-').replace('.', '').lower()
        topic_system_serial_number = str(self.system_serial_number).rjust(6, '0')

        return topic_model_name + '-' + topic_system_serial_number


    def message_tag(self):
        if self.vendor_id is None or self.model_id is None or self.system_serial_number is None:
            return None

        tag_vendor_id = self.vendor_id.lower()
        tag_model_id = self.model_id.lower()

        return tag_vendor_id + '-' + tag_model_id + '-' + str(self.system_serial_number)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        if self.last_modified:
            jdict['set-on'] = None if self.last_modified is None else self.last_modified.as_iso8601()

        jdict['vendor-id'] = self.vendor_id
        jdict['model-id'] = self.model_id

        jdict['model'] = self.model_name
        jdict['config'] = self.configuration

        jdict['system-sn'] = self.system_serial_number

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def vendor_id(self):
        return self.__vendor_id


    @property
    def model_id(self):
        return self.__model_id


    @property
    def model_name(self):
        return self.__model_name


    @property
    def configuration(self):
        return self.__configuration


    @property
    def system_serial_number(self):
        return self.__system_serial_number


    @system_serial_number.setter
    def system_serial_number(self, system_serial_number):
        self.__system_serial_number = system_serial_number


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "SystemID:{vendor_id:%s, model_id:%s, model_name:%s, configuration:%s, system_serial_number:%s}" % \
               (self.vendor_id, self.model_id, self.model_name, self.configuration, self.system_serial_number)



"""
Created on 30 Dec 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

Recommended Minimum data
$xxRMC,datetime,status,lat,NS,long,EW,spd,cog,date,mv,mv_ew,pos_mode*cs

example sentence:
$GPRMC,083559.00,A,4717.11437,N,00833.91522,E,0.004,77.52,091202,,,A*57
$GNRMC,103953.00,A,5049.38023,N,00007.38608,W,0.195,,310119,,,D*72

example values:
GPRMC:{datetime:GPDateTime:{date:110117, time:141101.00}, status:A,
        loc:GPLoc:{lat:5049.38464, ns:N, lng:00007.37785, ew:W}, spd:0.005, cog:None, mv:None, mv_ew:None, pos_mode:D}

GPRMC:{datetime:GPDateTime:{date:110117, time:140047.00}, status:V,
        loc:GPLoc:{lat:None, ns:None, lng:None, ew:None}, spd:None, cog:None, mv:None, mv_ew:None, pos_mode:N}

https://www.nmea.org
https://en.wikipedia.org/wiki/NMEA_0183
"""

from scs_core.position.nmea.gpdatetime import GPDateTime
from scs_core.position.nmea.gploc import GPLoc
from scs_core.position.nmea.nmea_sentence import NMEASentence


# --------------------------------------------------------------------------------------------------------------------

class GPRMC(NMEASentence):
    """
    classdocs
    """

    MESSAGE_IDS = ("$GNRMC", "$GPRMC")

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, r):
        if r.message_id not in cls.MESSAGE_IDS:
            raise TypeError("invalid sentence:%s" % r)

        time = r.str(1)
        status = r.str(2)

        lat = r.str(3)
        ns = r.str(4)

        lng = r.str(5)
        ew = r.str(6)

        loc = GPLoc(lat, ns, lng, ew)

        spd = r.float(7, 3)
        cog = r.float(8, 2)
        date = r.str(9)
        mv = r.float(10, 2)
        mv_ew = r.str(11)

        pos_mode = r.str(12)

        datetime = GPDateTime(date, time)

        return GPRMC(r.message_id, datetime, status, loc, spd, cog, mv, mv_ew, pos_mode)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, message_id, datetime, status, loc, spd, cog, mv, mv_ew, pos_mode):
        """
        Constructor
        """
        super().__init__(message_id)

        self.__datetime = datetime          # GPDateTime

        self.__status = status              # string
        self.__loc = loc                    # GPLoc

        self.__spd = spd                    # float(3) - speed over ground (knots)
        self.__cog = cog                    # float(2) - degrees course over ground
        self.__mv = mv                      # float(2) - degrees magnetic variation
        self.__mv_ew = mv_ew                # string - magnetic variation indicator

        self.__pos_mode = pos_mode          # string


    # ----------------------------------------------------------------------------------------------------------------

    def has_position(self):
        return self.__loc is not None and self.__loc.has_position()


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def datetime(self):
        return self.__datetime


    @property
    def status(self):
        return self.__status


    @property
    def loc(self):
        return self.__loc


    @property
    def spd(self):
        return self.__spd


    @property
    def cog(self):
        return self.__cog


    @property
    def mv(self):
        return self.__mv


    @property
    def mv_ew(self):
        return self.__mv_ew


    @property
    def pos_mode(self):
        return self.__pos_mode


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "GPRMC:{source:%s, datetime:%s, status:%s, loc:%s, spd:%s, cog:%s, mv:%s, mv_ew:%s, pos_mode:%s}" % \
                    (self.source, self.datetime, self.status, self.loc, self.spd, self.cog, self.mv, self.mv_ew,
                     self.pos_mode)



"""
Created on 1 Mar 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example JSON:
{"calibrated-on": "2021-06-02T13:11:31+01:00", "offset": -123,
"sample": {"rec": "2021-06-01T13:11:31+01:00", "hmd": 54.3, "tmp": 12.3}}
"""

from collections import OrderedDict

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.datum import Datum
from scs_core.data.json import JSONable
from scs_core.data.path_dict import PathDict


# --------------------------------------------------------------------------------------------------------------------

class SensorBaseline(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return cls.null_datum()

        if 'calibrated_on' in jdict:
            date = Datum.date(jdict.get('calibrated_on'))                   # deprecated
            calibrated_on = LocalizedDatetime.construct_from_date(date)

        else:
            calibrated_on = LocalizedDatetime.construct_from_iso8601(jdict.get('calibrated-on'))

        offset = jdict.get('offset')
        sample = SensorBaselineSample.construct_from_jdict(jdict.get('env'))

        return cls(calibrated_on, offset, sample=sample)


    @classmethod
    def null_datum(cls):
        return cls(None, 0)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, calibrated_on, offset, sample=None):
        """
        Constructor
        """
        self.__calibrated_on = calibrated_on                    # LocalizedDatetime
        self.__offset = int(offset)                             # int                           ppb
        self.__sample = sample                                  # SensorBaselineSample


    def __eq__(self, other):
        try:
            return self.calibrated_on == other.calibrated_on and self.offset == other.offset and \
                   self.sample == other.sample

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['calibrated-on'] = None if self.calibrated_on is None else self.calibrated_on.as_iso8601()
        jdict['offset'] = self.offset

        if self.sample is not None:
            jdict['env'] = self.sample

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def calibrated_on(self):
        return self.__calibrated_on


    @property
    def offset(self):
        return self.__offset


    @property
    def sample(self):
        return self.__sample


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "SensorBaseline:{calibrated_on:%s, offset:%s, sample:%s}" %  \
               (self.calibrated_on, self.offset, self.sample)


# --------------------------------------------------------------------------------------------------------------------

class SensorBaselineSample(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        rec = LocalizedDatetime.construct_from_iso8601(jdict.get('rec'))
        humid = jdict.get('hmd')
        temp = jdict.get('tmp')
        press = jdict.get('pA')

        return cls(rec, humid, temp, press)


    @classmethod
    def construct_from_sample_jdict(cls, jdict):
        if not jdict:
            return None

        sample = PathDict(jdict)

        rec = LocalizedDatetime.construct_from_iso8601(sample.node('rec'))
        humid = sample.node('val.sht.hmd')
        temp = sample.node('val.sht.tmp')
        press = sample.node('val.sht.bar.pA') if sample.has_path('val.sht.bar') else None

        return cls(rec, humid, temp, press)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, rec, humid, temp, press):
        """
        Constructor
        """
        self.__rec = rec                                # LocalizedDatetime
        self.__humid = Datum.float(humid, 1)            # float
        self.__temp = Datum.float(temp, 1)              # float
        self.__press = Datum.float(press, 1)            # float


    def __eq__(self, other):
        try:
            return self.rec == other.rec and self.humid == other.humid and self.temp == other.temp and \
                   self.press == other.press

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        if self.rec is not None:
            jdict['rec'] = self.rec.as_iso8601()

        jdict['hmd'] = self.humid
        jdict['tmp'] = self.temp

        if self.press is not None:
            jdict['pA'] = self.press

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def rec(self):
        return self.__rec


    @property
    def humid(self):
        return self.__humid


    @property
    def temp(self):
        return self.__temp


    @property
    def press(self):
        return self.__press


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "SensorBaselineSample:{rec:%s, humid:%s, temp:%s, press:%s}" %  \
               (self.rec, self.humid, self.temp, self.press)



"""
Created on 24 Jan 2024

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example request:
{"sample": {"rec": "2024-01-24T14:32:35Z", "tag": "scs-be2-3", "ver": 2.0, "src": "N3",
"val": {"per": 4.1, "pm1": 3.4, "pm2p5": 5.8, "pm10": 6.1,
"bin": [276, 175, 80, 13, 11, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
"mtf1": 27, "mtf3": 29, "mtf5": 40, "mtf7": 0, "sfr": 5.61,
"sht": {"hmd": 42.6, "tmp": 23.8}}},
"ext-sht": {"hmd": 46.6, "tmp": 24.9},
"slopes": {"ext-t": 0.1, "ext-rh": 0.2, "opc-t": 0.3, "opc-rh": 0.4}}
"""

import json

from scs_core.comms.uds_client import UDSClient

from scs_core.data.json import JSONify
from scs_core.data.linear_regression import LinearRegression

from scs_core.model.pmx.pmx_inference_client import PMxInferenceClient
from scs_core.model.pmx.o2.pmx_request import PMxRequest

from scs_core.sync.schedule import ScheduleItem

from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class O2PMxInferenceClient(PMxInferenceClient):
    """
    classdocs
    """

    @classmethod
    def construct(cls, socket, inference_uds_path, schedule_item: ScheduleItem):
        # UDS...
        uds_client = UDSClient(socket, inference_uds_path)

        # T / rH slope...
        slope_tally = PMxRequest.slope_tally(schedule_item.duration())

        meteo_t_regression = LinearRegression(tally=slope_tally)
        meteo_rh_regression = LinearRegression(tally=slope_tally)

        opc_t_regression = LinearRegression(tally=slope_tally)
        opc_rh_regression = LinearRegression(tally=slope_tally)

        return cls(uds_client, meteo_t_regression, meteo_rh_regression, opc_t_regression, opc_rh_regression)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, uds_client, meteo_t_regression, meteo_rh_regression, opc_t_regression, opc_rh_regression):
        """
        Constructor
        """
        super().__init__(uds_client)

        self.__meteo_t_regression = meteo_t_regression                      # LinearRegression
        self.__meteo_rh_regression = meteo_rh_regression                    # LinearRegression

        self.__opc_t_regression = opc_t_regression                          # LinearRegression
        self.__opc_rh_regression = opc_rh_regression                        # LinearRegression

        self.__logger = Logging.getLogger()


    # ----------------------------------------------------------------------------------------------------------------

    def infer(self, opc_sample, ext_sht_datum):
        # self.__logger.error("opc_sample: %s" % opc_sample)
        # self.__logger.error("ext_sht_datum: %s" % ext_sht_datum)

        # Meteo T / rH slope...
        self.__meteo_t_regression.append(opc_sample.rec, ext_sht_datum.temp)
        meteo_t_slope = self.__meteo_t_regression.slope(default=0.0)

        self.__meteo_rh_regression.append(opc_sample.rec, ext_sht_datum.humid)
        meteo_rh_slope = self.__meteo_rh_regression.slope(default=0.0)

        # OPC T / rH slope...
        self.__opc_t_regression.append(opc_sample.rec, opc_sample.opc_datum.sht.temp)
        opc_t_slope = self.__opc_t_regression.slope(default=0.0)

        self.__opc_rh_regression.append(opc_sample.rec, opc_sample.opc_datum.sht.humid)
        opc_rh_slope = self.__opc_rh_regression.slope(default=0.0)

        # request...
        pmx_request = PMxRequest(opc_sample, ext_sht_datum, meteo_t_slope, meteo_rh_slope, opc_t_slope, opc_rh_slope)
        # self.__logger.info("pmx_request: %s" % pmx_request)

        self._uds_client.request(JSONify.dumps(pmx_request.as_json()))
        response = self._uds_client.wait_for_response()

        return json.loads(response)


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "O2PMxInferenceClient:{uds_client:%s, meteo_t_regression:%s, meteo_rh_regression:%s, " \
               "opc_t_regression:%s, opc_rh_regression:%s}" %  \
            (self._uds_client, self.__meteo_t_regression, self.__meteo_rh_regression,
             self.__opc_t_regression, self.__opc_rh_regression)



"""
Created on 22 Dec 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

an abstract machine learning model group configuration
"""

import os

from abc import ABC, abstractmethod
from collections import OrderedDict

from scs_core.data.json import PersistentJSONable
from scs_core.model.model_map import ModelMap
from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class ModelConf(ABC, PersistentJSONable):
    """
    classdocs
    """

    __TEMPLATE_NO_MODELS = 'm0'

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def gg_ml_template(cls, gas_model_conf, pmx_model_conf):
        # no models...
        if not gas_model_conf and not pmx_model_conf:
            return cls.__TEMPLATE_NO_MODELS

        # PMx-only model...
        if not gas_model_conf and pmx_model_conf:
            return None if pmx_model_conf.model_map is None else pmx_model_conf.model_map.p_gg_ml_template

        # gas-only model...
        if gas_model_conf and not pmx_model_conf:
            raise ValueError('a gas model without a PMx model is not supported.')

        # gas and PMx models...
        gas_model_map = gas_model_conf.model_map
        pmx_model_map = pmx_model_conf.model_map

        if gas_model_map != pmx_model_map:
            raise ValueError("the gas model map '%s' does not match the PMx model map '%s'" %
                             (gas_model_map, pmx_model_map))

        return None if gas_model_conf.model_map is None else gas_model_conf.model_map.pg_gg_ml_template


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    @abstractmethod
    def interfaces(cls):
        return tuple()


    @classmethod
    def is_valid_interface(cls, model_interface):
        return model_interface in cls.interfaces()


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return cls(None, None, None) if skeleton else None

        uds_path = jdict.get('uds-path')
        model_interface = jdict.get('model-interface')

        try:
            field = 'model-map' if 'model-map' in jdict else 'model-compendium-group'
            model_map = ModelMap.map(jdict.get(field))
        except KeyError:
            model_map = None

        return cls(uds_path, model_interface, model_map)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, uds_path, model_interface, model_map):
        """
        Constructor
        """
        super().__init__()

        self.__uds_path = uds_path                                      # string
        self.__model_interface = model_interface                        # string
        self.__model_map = model_map                                    # ModelMapping

        self._logger = Logging.getLogger()


    def __eq__(self, other):
        try:
            return self.uds_path == other.uds_path and self.model_interface == other.model_interface and \
                   self.model_map == other.model_map

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['uds-path'] = self.uds_path
        jdict['model-interface'] = self.model_interface
        jdict['model-map'] = self.model_map_name

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def abs_uds_path(self, host):
        return os.path.join(host.scs_path(), self.uds_path)


    @property
    def uds_path(self):
        return self.__uds_path


    @property
    def model_interface(self):
        return self.__model_interface


    @property
    def model_map(self):
        return self.__model_map


    @property
    def model_map_name(self):
        return None if self.__model_map is None else self.__model_map.name


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        model_map_name = None if self.model_map is None else self.model_map.name

        return self.__class__.__name__ + ":{uds_path:%s, model_interface:%s, model_map:%s}" %  \
            (self.uds_path, self.model_interface, model_map_name)



"""
Created on 18 Sep 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

from collections import OrderedDict

from scs_core.climate.sht_datum import SHTDatum

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.datum import Datum

from scs_core.particulate.pmx_datum import PMxDatum

from scs_core.sample.sample import Sample

from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class OPCDatum(PMxDatum):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        logger = Logging.getLogger()

        source = jdict.get('src')

        rec = LocalizedDatetime.construct_from_jdict(jdict.get('rec'))

        period = jdict.get('per')

        pm1 = jdict.get('pm1')
        pm2p5 = jdict.get('pm2p5')
        pm10 = jdict.get('pm10')

        bins = jdict.get('bin')

        if bins is None:
            logger.error("OPCDatum incomplete: %s" % jdict)
            return None

        bin_1_mtof = jdict.get('mtf1')
        bin_3_mtof = jdict.get('mtf3')
        bin_5_mtof = jdict.get('mtf5')
        bin_7_mtof = jdict.get('mtf7')

        sfr = jdict.get('sfr')

        sht = SHTDatum.construct_from_jdict(jdict.get('sht'))

        return cls(source, rec, pm1, pm2p5, pm10, period, bins, bin_1_mtof, bin_3_mtof, bin_5_mtof, bin_7_mtof,
                   sfr=sfr, sht=sht)


    @classmethod
    def null_datum(cls):
        return cls(None, LocalizedDatetime.now().utc(), None, None, None, 0.0, [], 0, 0, 0, 0)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, source, rec, pm1, pm2p5, pm10, period, bins, bin_1_mtof, bin_3_mtof, bin_5_mtof, bin_7_mtof,
                 sfr=None, sht=None):
        """
        Constructor
        """
        PMxDatum.__init__(self, rec, pm1, pm2p5, None, pm10)

        self.__source = source                              # string

        self.__period = Datum.float(period, 1)              # seconds

        self.__bins = [int(count) for count in bins]        # array of count

        self.__bin_1_mtof = Datum.int(bin_1_mtof)           # float time
        self.__bin_3_mtof = Datum.int(bin_3_mtof)           # float time
        self.__bin_5_mtof = Datum.int(bin_5_mtof)           # float time
        self.__bin_7_mtof = Datum.int(bin_7_mtof)           # float time

        self.__sfr = Datum.float(sfr, 3)                    # float sample flow rate

        self.__sht = sht                                    # SHTDatum


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['src'] = self.source

        jdict['rec'] = None if self.rec is None else self.rec.as_iso8601(include_millis=Sample.INCLUDE_MILLIS)

        jdict['per'] = self.period

        jdict['pm1'] = self.pm1
        jdict['pm2p5'] = self.pm2p5
        jdict['pm10'] = self.pm10

        jdict['bin'] = self.bins

        jdict['mtf1'] = self.bin_1_mtof
        jdict['mtf3'] = self.bin_3_mtof
        jdict['mtf5'] = self.bin_5_mtof
        jdict['mtf7'] = self.bin_7_mtof

        jdict['sfr'] = self.sfr

        if self.sht is not None:
            jdict['sht'] = self.sht

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def is_zero(self):
        return self.pm1 == 0.0 and self.pm2p5 == 0.0 and self.pm10 == 0.0


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def source(self):
        return self.__source


    @property
    def period(self):
        return self.__period


    @property
    def bins(self):
        return self.__bins


    @property
    def bin_1_mtof(self):
        return self.__bin_1_mtof


    @property
    def bin_3_mtof(self):
        return self.__bin_3_mtof


    @property
    def bin_5_mtof(self):
        return self.__bin_5_mtof


    @property
    def bin_7_mtof(self):
        return self.__bin_7_mtof


    @property
    def sfr(self):
        return self.__sfr


    @property
    def sht(self):
        return self.__sht


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return self.__class__.__name__ + ":{source:%s, rec:%s, pm1:%s, pm2p5:%s, pm10:%s, period:%0.1f, bins:%s, " \
               "bin_1_mtof:%s, bin_3_mtof:%s, bin_5_mtof:%s, bin_7_mtof:%s, sfr:%s, sht:%s}" % \
               (self.source, self.rec, self.pm1, self.pm2p5, self.pm10, self.period, self.bins,
                self.bin_1_mtof, self.bin_3_mtof, self.bin_5_mtof, self.bin_7_mtof, self.sfr, self.sht)



"""
Created on 26 May 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import requests

from scs_core.aws.client.api_client import APIClient
from scs_core.aws.config.endpoint import APIEndpoint

from scs_core.aws.manager.configuration.configuration_check_requester_intercourse import \
    ConfigurationCheckRequesterResponse


# --------------------------------------------------------------------------------------------------------------------

class Endpoint(APIEndpoint):
    @classmethod
    def configuration(cls):
        return cls('ConfReqAPI/MQTTConfigQueuer',
                   'https://5nkrlhaq69.execute-api.us-west-2.amazonaws.com/default/MQTTConfigQueuer')


# --------------------------------------------------------------------------------------------------------------------

class ConfigurationCheckRequester(APIClient):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self):
        super().__init__()


    # ----------------------------------------------------------------------------------------------------------------

    def request(self, token, tag):
        params = {'tag': tag}

        response = requests.get(Endpoint.url(), headers=self._token_headers(token), params=params)
        self._check_response(response)

        return ConfigurationCheckRequesterResponse.construct_from_jdict(response.json())



"""
Created on 4 Mar 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import json

from abc import ABC, abstractmethod

from scs_core.csv.csv_reader import CSVReader


# --------------------------------------------------------------------------------------------------------------------

class CSVArchive(ABC):
    """
    classdocs
    """

    _retrieved = None                # must be overridden - as an {} - by each concrete class

    # ----------------------------------------------------------------------------------------------------------------

    # noinspection PyUnresolvedReferences

    @classmethod
    def retrieve(cls):
        reader = CSVReader.construct_for_file(cls.archive_location(), cast=False)

        try:
            for row in reader.rows():
                instance = cls.construct_from_jdict(json.loads(row))

                if instance.pk in cls._retrieved:
                    raise KeyError("duplicate pk '%s' for instance:%s" % (instance.pk, instance))

                cls._retrieved[instance.pk] = instance

        finally:
            reader.close()


    @classmethod
    def keys(cls):
        return tuple(cls._retrieved.keys())


    @classmethod
    def instances(cls):
        return tuple(cls._retrieved.values())


    @classmethod
    def instance(cls, pk):
        if pk is None:
            return None

        try:
            return cls._retrieved[pk]

        except KeyError:
            return None


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    @abstractmethod
    def archive_location(cls):
        return ''


    @classmethod
    @abstractmethod
    def construct_from_jdict(cls, _):
        return None


    # ----------------------------------------------------------------------------------------------------------------

    @property
    @abstractmethod
    def pk(self):
        return None



"""
Created on 23 Apr 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://docs.python.org/3/library/http.html
https://developer.mozilla.org/en-US/docs/Web/HTTP/CORS
"""

from abc import ABC
from http import HTTPStatus

from scs_core.client.http_exception import HTTPException
from scs_core.data.json import JSONable, JSONify


# --------------------------------------------------------------------------------------------------------------------

class HTTPResponse(JSONable, ABC):
    """
    classdocs
    """

    __CORS_HEADERS = {                                  # Cross-Origin Resource Sharing
        'Access-Control-Allow-Origin': '*',
        'Access-Control-Allow-Credentials': True,
    }


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_res(cls, response):         # response: requests.Response
        status = HTTPStatus(response.status_code)

        if status != HTTPStatus.OK:
            raise HTTPException.construct(status.value, response.reason, response.json())

        jdict = response.json()

        if not jdict:
            return None

        return cls.construct_from_response_jdict(status, jdict)


    @classmethod
    def construct_from_response_jdict(cls, status, jdict):
        return cls(status, body=jdict)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, status, body=None):
        """
        Constructor
        """
        self.__status = status                          # HTTPStatus member
        self.__body = body                              # string


    def __len__(self):
        try:
            return len(self.body)
        except (AttributeError, TypeError):
            return 1


    # ----------------------------------------------------------------------------------------------------------------

    def is_ok(self):
        return self.status == HTTPStatus.OK


    # ----------------------------------------------------------------------------------------------------------------

    def as_http(self, cors=False):
        jdict = {
            'statusCode': int(self.status),
            'body': JSONify.dumps(self)
        }

        if cors:
            jdict['headers'] = self.__CORS_HEADERS

        return jdict


    def as_json(self):
        return self.body


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def status(self):
        return self.__status


    @property
    def body(self):
        return self.__body


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "HTTPResponse:{status:%s, body:%s}" % (self.status, self.body)



"""
Created on 28 Apr 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import requests

from scs_core.aws.client.api_client import APIClient
from scs_core.aws.config.endpoint import APIEndpoint

from scs_core.aws.manager.configuration.configuration_check_intercourse import ConfigurationCheckRequest, \
    ConfigurationCheckResponse


# --------------------------------------------------------------------------------------------------------------------

class Endpoint(APIEndpoint):
    @classmethod
    def configuration(cls):
        return cls('ConfChkAPI/ConfigurationCheckFinder',
                   'https://03azvwwip4.execute-api.us-west-2.amazonaws.com/default/ConfigurationCheckFinder')


# --------------------------------------------------------------------------------------------------------------------

class ConfigurationCheckFinder(APIClient):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self):
        super().__init__()


    # ----------------------------------------------------------------------------------------------------------------

    def find(self, token, tag_filter, exact_match, response_mode):
        request = ConfigurationCheckRequest(tag_filter, exact_match, response_mode)

        response = requests.get(Endpoint.url(), headers=self._token_headers(token), params=request.params())
        self._check_response(response)

        return ConfigurationCheckResponse.construct_from_jdict(response.json())



"""
Created on 26 Aug 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

from collections import OrderedDict
from enum import Enum

from scs_core.data.json import JSONReport


# --------------------------------------------------------------------------------------------------------------------

class QueueReport(JSONReport):
    """
    classdocs
   """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return QueueReport(0, ClientStatus.WAITING, False)

        length = jdict.get('length')
        client_state = ClientStatus[jdict.get('client-state')]
        publish_success = jdict.get('publish-success')

        return QueueReport(length, client_state, publish_success)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, length, client_state, publish_success):
        """
        Constructor
        """
        self.__length = length                                  # int or None
        self.__client_state = client_state                      # int
        self.__publish_success = publish_success                # bool


    # ----------------------------------------------------------------------------------------------------------------

    def queue_state(self):
        # client INHIBITED...
        if self.client_state == ClientStatus.INHIBITED:
            return QueueStatus.INHIBITED

        # client WAITING...
        if self.client_state == ClientStatus.WAITING:
            return QueueStatus.STARTING

        # client CONNECTING...
        if self.client_state == ClientStatus.CONNECTING:
            return QueueStatus.CONNECTING

        # client CONNECTED...
        if self.client_state == ClientStatus.CONNECTED:
            if self.length == 0:
                return QueueStatus.WAITING_FOR_DATA

            if self.publish_success:
                return QueueStatus.PUBLISHING

            return QueueStatus.QUEUING

        # unknown / error...
        return QueueStatus.NONE


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['length'] = self.length
        jdict['client-state'] = self.client_state.name
        jdict['publish-success'] = self.publish_success

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def length(self):
        return self.__length


    @length.setter
    def length(self, length):
        self.__length = length


    @property
    def client_state(self):
        return self.__client_state


    @client_state.setter
    def client_state(self, client_state):
        self.__client_state = client_state


    @property
    def publish_success(self):
        return self.__publish_success


    @publish_success.setter
    def publish_success(self, publish_success):
        self.__publish_success = publish_success


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "QueueReport:{length:%s, client_state:%s, publish_success:%s}" % \
               (self.length, self.client_state, self.publish_success)


# --------------------------------------------------------------------------------------------------------------------

class ClientStatus(Enum):
    NONE =              1
    INHIBITED =         2
    WAITING =           3
    CONNECTING =        4
    CONNECTED =         5


# --------------------------------------------------------------------------------------------------------------------

class QueueStatus(Enum):
    NONE =              1
    INHIBITED =         2
    STARTING =          3
    CONNECTING =        4
    WAITING_FOR_DATA =  5
    PUBLISHING =        6
    QUEUING =           7
    CLEARING =          8



"""
Created on 30 Sep 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://www.alphasense-technology.co.uk/sensor_types
"""

from abc import ABC, abstractmethod


# --------------------------------------------------------------------------------------------------------------------

class Sensor(ABC):
    """
    classdocs
    """

    CODE_CO =           '132'           # CO A4
    CODE_H2S =          '133'           # H2SA4
    CODE_NO =           '130'           # NO A4
    CODE_NO2 =          '212'           # NOGA4, NO2A43F
    CODE_OX =           '214'           # OXGA4, OXA431
    CODE_SO2 =          '134'           # SO2A4

    CODE_VOCe =         '217'           # VOCA4

    CODE_VOC_PPM =      '142'           # PID12
    CODE_VOC_PPB_T1 =   '143'           # PIDH2
    CODE_VOC_PPB_T2 =   '354'           # PID-AH (since June 2019)
    CODE_VOC_PPB_T3 =   '401'           # PID-AH (since April 2020)
    CODE_VOC_PPB_T4 =   '452'           # PID-AH (since May 2021)
    CODE_VOC_PPB_T5 =   '507'           # PID-AH Rev. 2 (since May 2021)

    CODE_TEST_1 =       '01'            # test load
    CODE_TEST_2 =       '02'            # test load
    CODE_TEST_3 =       '03'            # test load
    CODE_TEST_4 =       '04'            # test load

    SENSORS =       {}

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def find(cls, serial_number):
        if serial_number is None:
            return None

        for code, sensor in cls.SENSORS.items():
            if str(serial_number).startswith(code):
                return sensor

        return None


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, sensor_code, sensor_type, gas_name, adc_gain_index):
        """
        Constructor
        """
        self.__sensor_code = sensor_code
        self.__sensor_type = sensor_type

        self.__gas_name = gas_name
        self.__adc_gain_index = adc_gain_index

        self._calib = None
        self._calibrator = None
        self.__baseline = None


    # ----------------------------------------------------------------------------------------------------------------

    def has_no2_cross_sensitivity(self):
        if self.calib is None:
            return False

        return self.calib.has_no2_cross_sensitivity()


    @abstractmethod
    def sample(self, afe, temp, index, no2_sample=None):
        pass


    @abstractmethod
    def null_datum(self):
        pass


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def sensor_code(self):
        return self.__sensor_code


    @property
    def sensor_type(self):
        return self.__sensor_type


    @property
    def gas_name(self):
        return self.__gas_name


    @property
    def adc_gain_index(self):
        return self.__adc_gain_index


    # ----------------------------------------------------------------------------------------------------------------

    @property
    @abstractmethod
    def calib(self):
        pass


    @calib.setter
    @abstractmethod
    def calib(self, calib):
        pass


    @property
    def calibrator(self):
        return self._calibrator


    @property
    def baseline(self):
        return self.__baseline


    @baseline.setter
    def baseline(self, baseline):
        self.__baseline = baseline



"""
Created on 22 Aug 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

If tally is None: computes the average of all the appended data.
If tally is a positive integer N: computes the average of the last N appended data.
"""


# --------------------------------------------------------------------------------------------------------------------

class Average(object):
    """
    classdocs
    """

    MIN_DATA_POINTS =   1

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, tally=None):
        """
        Constructor
        """
        self.__tally = tally                            # number of rolling samples (None for all samples)
        self.__data = []


    def __len__(self):
        return len(self.__data)


    # ----------------------------------------------------------------------------------------------------------------

    def has_midpoint(self):
        return len(self.__data) > 0


    def has_tally(self):
        count = len(self)

        if self.__tally is None:
            return count >= self.MIN_DATA_POINTS

        return count >= self.__tally


    def append(self, value):
        count = len(self.__data)

        # remove oldest?
        if self.__tally is not None and count == self.__tally:
            del self.__data[0]

        # append...
        self.__data.append(value)


    def reset(self):
        self.__data = []


    # ----------------------------------------------------------------------------------------------------------------

    def mid(self, ndigits=None):
        count = len(self)

        if count < Average.MIN_DATA_POINTS:
            return None

        total = None
        for value in self.__data:
            total = value if total is None else total + value

        average = total / count

        return average if ndigits is None else round(average, ndigits)


    # ----------------------------------------------------------------------------------------------------------------

    def min(self):
        return min(self.__data)


    def max(self):
        return max(self.__data)


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Average:{tally:%s, items:%s}" % (self.__tally, len(self))



"""
Created on 22 Feb 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://www.w3schools.com/sql/sql_join.asp
"""

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.path_dict import PathDict


# --------------------------------------------------------------------------------------------------------------------

class Join(object):
    """
    classdocs
    """

    TYPES = {'INNER', 'LEFT', 'RIGHT', 'FULL'}

    @classmethod
    def is_valid_type(cls, name):
        return name.upper() in cls.TYPES


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, left_set_path, left_pk_path, right_set_path, right_pk_path, pk_is_iso8601):
        left = JoinSet(left_set_path, left_pk_path, pk_is_iso8601)
        right = JoinSet(right_set_path, right_pk_path, pk_is_iso8601)

        return Join(left, right)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, left, right):
        """
        Constructor
        """
        self.__left = left                      # JoinSet
        self.__right = right                    # JoinSet


    # ----------------------------------------------------------------------------------------------------------------

    def append_to_left(self, document: PathDict):
        self.__left.append(document)


    def append_to_right(self, document: PathDict):
        self.__right.append(document)


    # ----------------------------------------------------------------------------------------------------------------

    def inner(self):
        # paths...
        pk_path = self.__left.pk_path
        right_path = self.__right.set_path
        left_path = self.__left.set_path

        # join...
        for pk in self.__left.pk_values():
            right = self.__right.retrieve(pk)

            if right is None:
                continue

            left = self.__left.retrieve(pk)

            yield PathDict.union((pk_path, pk), (left_path, left), (right_path, right))


    def left(self):
        # paths...
        pk_path = self.__left.pk_path
        right_path = self.__right.set_path
        left_path = self.__left.set_path

        # join...
        for pk in self.__left.pk_values():
            right = self.__right.retrieve(pk)
            left = self.__left.retrieve(pk)

            yield PathDict.union((pk_path, pk), (left_path, left), (right_path, right))


    def right(self):
        # paths...
        pk_path = self.__right.pk_path
        right_path = self.__right.set_path
        left_path = self.__left.set_path

        # join...
        for pk in self.__right.pk_values():
            right = self.__right.retrieve(pk)
            left = self.__left.retrieve(pk)

            yield PathDict.union((pk_path, pk), (left_path, left), (right_path, right))


    def full(self):
        # paths...
        pk_path = self.__left.pk_path
        right_path = self.__right.set_path
        left_path = self.__left.set_path

        # keys...
        right_pk_values = set(self.__right.pk_values())
        left_pk_values = set(self.__left.pk_values())

        pk_values = sorted(right_pk_values | left_pk_values)

        # join...
        for pk in pk_values:
            right = self.__right.retrieve(pk)
            left = self.__left.retrieve(pk)

            yield PathDict.union((pk_path, pk), (left_path, left), (right_path, right))


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Join:{left:%s, right:%s}" % (self.__left, self.__right)


# --------------------------------------------------------------------------------------------------------------------

class JoinSet(object):
    """
    classdocs
    """

    def __init__(self, set_path, pk_path, pk_is_iso8601):
        """
        Constructor
        """
        self.__set_path = set_path
        self.__pk_path = pk_path
        self.__pk_is_iso8601 = pk_is_iso8601

        self.__documents = {}


    def __len__(self):
        return len(self.__documents)


    # ----------------------------------------------------------------------------------------------------------------

    def append(self, document: PathDict):
        pk_value = document.node(self.pk_path)

        if self.pk_is_iso8601:
            datetime = LocalizedDatetime.construct_from_iso8601(pk_value)

            if datetime is None:
                raise ValueError(pk_value)

            pk_value = datetime

        self.__documents[pk_value] = document


    # ----------------------------------------------------------------------------------------------------------------

    def pk_values(self):
        for pk_value in self.__documents.keys():
            yield pk_value


    def retrieve(self, pk_value):
        try:
            document = self.__documents[pk_value]
        except KeyError:
            return None

        node = PathDict()

        for path in document.paths():
            if path != self.pk_path:
                node.append(path, document.node(path))

        return node


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def set_path(self):
        return self.__set_path


    @property
    def pk_path(self):
        return self.__pk_path


    @property
    def pk_is_iso8601(self):
        return self.__pk_is_iso8601


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "JoinSet:{set_path:%s, pk_path:%s, pk_is_iso8601:%s, len:%d}" % \
               (self.set_path, self.pk_path, self.pk_is_iso8601, len(self))



"""
Created on 24 Apr 2023

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

Enables a device to register itself as a CognitoDevice
"""

import requests

from scs_core.aws.client.api_client import APIClient
from scs_core.aws.config.endpoint import APIEndpoint
from scs_core.aws.security.cognito_device import CognitoDeviceIdentity

from scs_core.data.json import JSONify


# --------------------------------------------------------------------------------------------------------------------

class Endpoint(APIEndpoint):
    @classmethod
    def configuration(cls):
        return cls('CogDevCreAPI/CognitoDeviceCreator',
                   'https://bf32xbgymi.execute-api.us-west-2.amazonaws.com/default/CognitoDeviceCreator')


# --------------------------------------------------------------------------------------------------------------------

class CognitoDeviceCreator(APIClient):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self):
        super().__init__()


    # ----------------------------------------------------------------------------------------------------------------

    def may_create(self, device_tag):
        payload = {"DeviceTag": device_tag}

        response = requests.get(Endpoint.url(), headers=self._auth_headers(), data=JSONify.dumps(payload))
        self._check_response(response)

        return response.json()


    def create(self, identity: CognitoDeviceIdentity):
        response = requests.post(Endpoint.url(), headers=self._auth_headers(), data=JSONify.dumps(identity))
        self._check_response(response)

        return CognitoDeviceIdentity.construct_from_jdict(response.json())



"""
Created on 2 Oct 2023

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

Equivalent to cURLs:
curl "https://aws.southcoastscience.com/device-topics?topic=south-coast-science-dev/alphasense/loc/303/gases"
curl "https://aws.southcoastscience.com/device-topics?device=scs-bgx-303"
"""

from collections import OrderedDict
from urllib.parse import parse_qs, urlparse

from scs_core.aws.client.api_intercourse import APIResponse
from scs_core.aws.manager.byline.byline import Byline

from scs_core.data.str import Str


# --------------------------------------------------------------------------------------------------------------------

class BylineFinderResponse(APIResponse):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        items = []
        if jdict.get('Items'):
            for item_jdict in jdict.get('Items'):
                item = Byline.construct_from_jdict(item_jdict)
                items.append(item)

        next_url = jdict.get('next')

        return cls(items, next_url=next_url)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, items, next_url=None):
        """
        Constructor
        """
        self.__items = items                                # list of Byline
        self.__next_url = next_url                          # URL string


    def __len__(self):
        return len(self.items)


    # ----------------------------------------------------------------------------------------------------------------

    def next_params(self, _):
        return parse_qs(urlparse(self.next_url).query)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        if self.items is not None:
            jdict['Items'] = self.items
            jdict['itemCount'] = len(self.items)

        if self.next_url is not None:
            jdict['next'] = self.next_url

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def items(self):
        return self.__items


    @property
    def next_url(self):
        return self.__next_url


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "BylineFinderResponse:{items:%s, next_url:%s}" %  (Str.collection(self.items), self.next_url)



"""
Created on 1 Aug 2023

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import requests

from scs_core.aws.client.api_client import APIClient
from scs_core.aws.manager.byline.byline import Byline, DeviceBylineGroup, TopicBylineGroup
from scs_core.aws.manager.byline.byline_intercourse import BylineFinderResponse
from scs_core.aws.manager.byline.byline_manager import Endpoint


# --------------------------------------------------------------------------------------------------------------------

class BylineFinder(APIClient):
    """
    classdocs
    """

    DEVICE = 'device'
    TOPIC = 'topic'
    INCLUDE_MESSAGES = 'include-messages'

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, reporter=None):
        super().__init__(reporter=reporter)


    # ----------------------------------------------------------------------------------------------------------------

    def find_latest_byline_for_topic(self, token, topic, include_messages=True):
        params = {self.TOPIC: topic, self.INCLUDE_MESSAGES: include_messages}

        response = requests.get(Endpoint.url(), headers=self._token_headers(token), params=params)
        jdict = response.json()

        # bylines...
        if jdict is None:
            return None

        latest_byline = None

        for item in jdict:
            byline = Byline.construct_from_jdict(item)

            if latest_byline is None or latest_byline.rec < byline.rec:
                latest_byline = byline

        return latest_byline


    def find_bylines(self, token, excluded=None, strict_tags=False, include_messages=True):
        params = {self.INCLUDE_MESSAGES: include_messages}

        bylines = [item for item in self._get_blocks(Endpoint.url(), token, BylineFinderResponse, params=params)]

        return TopicBylineGroup.construct(bylines, excluded=excluded, strict_tags=strict_tags)


    def find_bylines_for_topic(self, token, topic, excluded=None, strict_tags=False, include_messages=True):
        params = {self.TOPIC: topic, self.INCLUDE_MESSAGES: include_messages}

        response = requests.get(Endpoint.url(), headers=self._token_headers(token), params=params)
        self._check_response(response)

        jdict = response.json()

        # bylines...
        return TopicBylineGroup.construct_from_jdict(jdict, excluded=excluded, strict_tags=strict_tags, skeleton=True)


    def find_bylines_for_device(self, token, device, excluded=None, include_messages=True):
        params = {self.DEVICE: device, self.INCLUDE_MESSAGES: include_messages}

        response = requests.get(Endpoint.url(), headers=self._token_headers(token), params=params)
        self._check_response(response)

        jdict = response.json()

        # bylines...
        return DeviceBylineGroup.construct_from_jdict(jdict, excluded=excluded, skeleton=True)


# --------------------------------------------------------------------------------------------------------------------

class DeviceBylineFinder(APIClient):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self):
        super().__init__()


    # ----------------------------------------------------------------------------------------------------------------

    def find_byline_for_topic(self, token, topic, include_messages=True):
        url = Endpoint.url('self')
        params = {BylineFinder.INCLUDE_MESSAGES: include_messages}

        response = requests.get(url, headers=self._token_headers(token), params=params)
        self._check_response(response)

        jdict = response.json()

        # bylines...
        if jdict is None:
            return None

        for item in jdict:
            byline = Byline.construct_from_jdict(item)

            if byline.topic == topic:
                return byline

        return None



"""
Created on 26 Oct 2021

@author: Jade Page (jade.page@southcoastscience.com)

The foundation data classes for the security system

example Organisation:
{"OrgID": 1, "Label": "SCS", "LongName": "South Coast Science", "URL": "https://www.southcoastscience.com",
"Owner": "bruno.beloff@southcoastscience.com", "ParentID": null}

example OrganisationPathRoot:
{"OPRID": 11, "OrgID": 1, "PathRoot": "ricardo/"}

example OrganisationUser:
{"Username": 111, "OrgID": 1, "IsOrgAdmin": true, "IsDeviceAdmin": true, "IsSuspended": false}

example OrganisationUserPath:
{"Username": 111, "OPRID": 11, "PathExtension": "heathrow/"}

example OrganisationDevice:
{"DeviceTag": "scs-bgx-401", "OrgID": 1, "DevicePath": "south-coast-science-demo/brighton/loc/1/",
"EnvironmentPath": "south-coast-science-demo/brighton/device/praxis-000401/",
"StartDatetime": "2022-01-17T10:40:04Z", "EndDatetime": null,
"DeploymentLabel": "Preston Circus"}
"""

import re

from collections import OrderedDict

from scs_core.aws.security.cognito_device import CognitoDeviceCredentials

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.datum import Datum
from scs_core.data.json import JSONable
from scs_core.data.timedelta import Timedelta


# --------------------------------------------------------------------------------------------------------------------

class Organisation(JSONable):
    """
    classdocs
    """

    ORG_ID = 'OrgID'
    LABEL = 'Label'
    LONG_NAME = 'LongName'
    URL = 'URL'
    OWNER = 'Owner'
    PARENT = 'ParentID'

    __EXCLUDED_EMAILS = ['paul@p3d.co.uk']
    __EXCLUDED_DOMAINS = ['southcoastscience.com']
    __UNFILTERED_ORGANISATIONS = ['South Coast Science']

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def is_valid_label(cls, label):
        try:
            if not (1 < len(label) < 256):
                return False

            return bool(re.fullmatch(r'[0-9A-Za-z\- /&.()]+', label))

        except TypeError:
            return False


    @classmethod
    def is_valid_long_name(cls, long_name):
        try:
            return 1 < len(long_name) < 256

        except TypeError:
            return False


    @classmethod
    def is_valid_url(cls, url):
        try:
            if len(url) > 255:
                return False

            return Datum.is_url(url)

        except TypeError:
            return False


    @classmethod
    def is_valid_owner(cls, owner):
        try:
            if len(owner) > 255:
                return False

            return Datum.is_email_address(owner)

        except TypeError:
            return False


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        org_id = jdict.get(cls.ORG_ID)
        label = jdict.get(cls.LABEL)
        long_name = jdict.get(cls.LONG_NAME)
        url = jdict.get(cls.URL)
        owner = jdict.get(cls.OWNER)
        parent_id = jdict.get(cls.PARENT)

        return cls(org_id, label, long_name, url, owner, parent_id)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, org_id, label, long_name, url, owner, parent_id):
        """
        Constructor
        """
        self._org_id = Datum.int(org_id)                # AUTO PK: int
        self.__label = label                            # UNIQUE: string
        self.__long_name = long_name                    # string
        self.__url = url                                # string
        self.__owner = owner                            # INDEX: string (email address)
        self.__parent_id = Datum.int(parent_id)         # FK: int (self reference)


    def __eq__(self, other):
        try:
            return self.label == other.label and self.long_name == other.long_name \
                   and self.url == other.url and self.owner == other.owner

        except (TypeError, AttributeError):
            return False


    def __lt__(self, other):
        return self.label.lower() < other.label.lower()


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):         # WARNING: does not test for label uniqueness
        return self.is_valid_label(self.label) and self.is_valid_long_name(self.long_name) and \
               self.is_valid_url(self.url) and self.is_valid_owner(self.owner)


    def organisation_path_root(self, aws_opr_id, path_root):
        return OrganisationPathRoot(aws_opr_id, self.org_id, path_root)


    def filtered_users(self, user_emails):
        for label_prefix in self.__UNFILTERED_ORGANISATIONS:
            if self.label.startswith(label_prefix):
                return user_emails

        filtered_users = []

        for user_email in user_emails:
            if user_email in self.__EXCLUDED_EMAILS:
                continue

            if user_email.split('@')[1] in self.__EXCLUDED_DOMAINS:
                continue

            filtered_users.append(user_email)

        return filtered_users


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict[self.ORG_ID] = self.org_id
        jdict[self.LABEL] = self.label
        jdict[self.LONG_NAME] = self.long_name
        jdict[self.URL] = self.url
        jdict[self.OWNER] = self.owner
        jdict[self.PARENT] = self.parent_id

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def org_id(self):
        return self._org_id


    @property
    def label(self):
        return self.__label


    @property
    def long_name(self):
        return self.__long_name


    @property
    def url(self):
        return self.__url


    @property
    def owner(self):
        return self.__owner


    @property
    def parent_id(self):
        return self.__parent_id


    # ------------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Organisation:{org_id:%s, label:%s, long_name:%s, url:%s, owner:%s, parent_id:%s}" % \
            (self.org_id, self.label, self.long_name, self.url, self.owner, self.parent_id)


# --------------------------------------------------------------------------------------------------------------------

class OrganisationPathRoot(JSONable):
    """
    classdocs
    """

    OPR_ID = 'OPRID'
    ORG_ID = 'OrgID'
    PATH_ROOT = 'PathRoot'

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def is_valid_path_root(cls, path_root):
        try:
            if len(path_root) > 255:
                return False

            return bool(re.fullmatch(r'[0-9A-Za-z\-]+/', path_root))

        except TypeError:
            return False


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        opr_id = jdict.get(cls.OPR_ID, 0)
        org_id = jdict.get(cls.ORG_ID)
        path_root = jdict.get(cls.PATH_ROOT)

        return cls(opr_id, org_id, path_root)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, opr_id, org_id, path_root):
        """
        Constructor
        """
        self._opr_id = Datum.int(opr_id)                # AUTO PK: int
        self.__org_id = Datum.int(org_id)               # FK: int
        self.__path_root = path_root                    # UNIQUE: string


    def __eq__(self, other):
        try:
            return self.org_id == other.org_id and self.path_root == other.path_root

        except (TypeError, AttributeError):
            return False


    def __lt__(self, other):
        return self.path_root < other.path_root


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def index(self):
        return self.path_root


    def is_valid(self):             # WARNING: does not test for path root uniqueness
        if self.org_id is None:
            return False

        return self.is_valid_path_root(self.path_root)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict[self.OPR_ID] = self.opr_id
        jdict[self.ORG_ID] = self.org_id
        jdict[self.PATH_ROOT] = self.path_root

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def opr_id(self):
        return self._opr_id


    @opr_id.setter
    def opr_id(self, value):
        self._opr_id = value


    @property
    def org_id(self):
        return self.__org_id


    @property
    def path_root(self):
        return self.__path_root

    # ------------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "OrganisationPathRoot:{opr_id:%s, org_id:%s, path_root:%s}" % \
            (self.opr_id, self.org_id, self.path_root)


# --------------------------------------------------------------------------------------------------------------------

class OrganisationUser(JSONable):
    """
    classdocs
    """

    USERNAME = 'Username'
    ORG_ID = 'OrgID'
    IS_ORG_ADMIN = 'IsOrgAdmin'
    IS_DEVICE_ADMIN = 'IsDeviceAdmin'
    IS_SUSPENDED = 'IsSuspended'

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        username = jdict.get(cls.USERNAME)
        org_id = jdict.get(cls.ORG_ID)
        is_org_admin = jdict.get(cls.IS_ORG_ADMIN)
        is_device_admin = jdict.get(cls.IS_DEVICE_ADMIN)
        is_suspended = jdict.get(cls.IS_SUSPENDED)

        return cls(username, org_id, is_org_admin, is_device_admin, is_suspended)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, username, org_id, is_org_admin, is_device_admin, is_suspended):
        """
        Constructor
        """
        self._username = username                                   # PK: string
        self._org_id = Datum.int(org_id)                            # PK, FK: int
        self.__is_org_admin = bool(is_org_admin)                    # INDEX: bool
        self.__is_device_admin = bool(is_device_admin)              # INDEX: bool
        self.__is_suspended = bool(is_suspended)                    # INDEX: bool


    def __eq__(self, other):
        try:
            return self.username == other.username and self.org_id == other.org_id \
                   and self.is_org_admin == other.is_org_admin and self.is_device_admin == other.is_device_admin \
                   and self.is_suspended == other.is_suspended

        except (TypeError, AttributeError):
            return False


    def __lt__(self, other):                    # requires join with Cognito to do useful sort
        if self.username < other.username:
            return True

        if self.username > other.username:
            return False

        return self.org_id < other.org_id


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        if self.username is None or self.org_id is None:
            return False

        return True


    def as_json(self):
        jdict = OrderedDict()

        jdict[self.USERNAME] = self.username
        jdict[self.ORG_ID] = self.org_id
        jdict[self.IS_ORG_ADMIN] = self.is_org_admin
        jdict[self.IS_DEVICE_ADMIN] = self.is_device_admin
        jdict[self.IS_SUSPENDED] = self.is_suspended

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def pk(self):
        return '+'.join((str(self.username), str(self.org_id)))


    @property
    def username(self):
        return self._username


    @property
    def org_id(self):
        return self._org_id


    @property
    def is_org_admin(self):
        return self.__is_org_admin


    @is_org_admin.setter
    def is_org_admin(self, is_org_admin):
        self.__is_org_admin = is_org_admin


    @property
    def is_device_admin(self):
        return self.__is_device_admin


    @property
    def is_suspended(self):
        return self.__is_suspended


    # ------------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "OrganisationUser:{username:%s, org_id:%s, is_org_admin:%s, is_device_admin:%s, is_suspended:%s}" % \
            (self.username, self.org_id, self.is_org_admin, self.is_device_admin, self.is_suspended)


# --------------------------------------------------------------------------------------------------------------------

class OrganisationUserPath(JSONable):
    """
    classdocs
    """

    USERNAME = 'Username'
    OPR_ID = 'OPRID'
    PATH_EXTENSION = 'PathExtension'

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def is_valid_path_extension(cls, path_extension):
        try:
            if len(path_extension) > 255:
                return False

            return bool(re.fullmatch(r'([0-9A-Za-z\-]+/)+', path_extension))

        except TypeError:
            return False


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        username = jdict.get(cls.USERNAME)
        opr_id = jdict.get(cls.OPR_ID)
        path_extension = jdict.get(cls.PATH_EXTENSION)

        return cls(username, opr_id, path_extension)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, username, opr_id, path_extension):
        """
        Constructor
        """
        self._username = username                                   # PK: string
        self._opr_id = Datum.int(opr_id)                            # PK, FK: int
        self._path_extension = path_extension                       # PK: string


    def __eq__(self, other):
        try:
            return self.username == other.username and self.opr_id == other.opr_id \
                   and self.path_extension == other.path_extension

        except (TypeError, AttributeError):
            return False


    def __lt__(self, other):
        if self.username < other.username:
            return True

        if self.username > other.username:
            return False

        if self.opr_id < other.opr_id:
            return True

        if self.opr_id > other.opr_id:
            return False

        return self.path_extension < other.path_extension


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        if self.username is None or self.opr_id is None:
            return False

        return self.is_valid_path_extension(self.path_extension)


    def as_json(self):
        jdict = OrderedDict()

        jdict[self.USERNAME] = self.username
        jdict[self.OPR_ID] = self.opr_id
        jdict[self.PATH_EXTENSION] = self.path_extension

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def username(self):
        return self._username


    @property
    def opr_id(self):
        return self._opr_id


    @property
    def path_extension(self):
        return self._path_extension


    # ------------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "OrganisationUserPath:{username:%s, opr_id:%s, path_extension:%s}" %  \
            (self.username, self.opr_id, self.path_extension)


# --------------------------------------------------------------------------------------------------------------------

class OrganisationDevice(JSONable):
    """
    classdocs
    """

    DEVICE_TAG = 'DeviceTag'
    ORG_ID = 'OrgID'
    DEVICE_PATH = 'DevicePath'
    LOCATION_PATH = 'LocationPath'
    START_DATETIME = 'StartDatetime'
    END_DATETIME = 'EndDatetime'
    DEPLOYMENT_LABEL = 'DeploymentLabel'

    # ----------------------------------------------------------------------------------------------------------------

    __DEPLOYMENT_INTERVAL = Timedelta(minutes=10)

    @classmethod
    def deployment_interval(cls):
        return cls.__DEPLOYMENT_INTERVAL


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def is_valid_deployment_label(cls, deployment_label):
        try:
            return 1 < len(deployment_label) < 256

        except TypeError:
            return False


    @classmethod
    def is_valid_path(cls, path):
        try:
            if len(path) > 255:
                return False

            return bool(re.fullmatch(r'([0-9A-Za-z\-]+/)+', path))

        except TypeError:
            return False


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        device_tag = jdict.get(cls.DEVICE_TAG)
        org_id = jdict.get(cls.ORG_ID)
        deployment_label = jdict.get(cls.DEPLOYMENT_LABEL)
        device_path = jdict.get(cls.DEVICE_PATH)
        location_path = jdict.get(cls.LOCATION_PATH)

        start_datetime = LocalizedDatetime.construct_from_iso8601(jdict.get(cls.START_DATETIME))
        end_datetime = LocalizedDatetime.construct_from_iso8601(jdict.get(cls.END_DATETIME))

        return cls(device_tag, org_id, device_path, location_path, start_datetime, end_datetime, deployment_label)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, device_tag, org_id, device_path, location_path, start_datetime, end_datetime,
                 deployment_label):
        """
        Constructor
        """
        self._device_tag = device_tag                   # PK: string
        self._org_id = Datum.int(org_id)                # PK, FK: int
        self.__device_path = device_path                # string
        self.__location_path = location_path            # string

        self._start_datetime = start_datetime           # PK: LocalizedDatetime
        self.__end_datetime = end_datetime              # LocalizedDatetime
        self.__deployment_label = deployment_label      # INDEX: string


    def __eq__(self, other):
        try:
            return self.device_tag == other.device_tag and self.org_id == other.org_id \
                and self.device_path == other.device_path and self.location_path == other.location_path \
                and self.start_datetime == other.start_datetime and self.end_datetime == other.end_datetime \
                and self.deployment_label == other.deployment_label

        except (TypeError, AttributeError):
            return False


    def __lt__(self, other):
        if self.device_tag < other.device_tag:
            return True

        if self.device_tag > other.device_tag:
            return False

        if self.start_datetime < other.start_datetime:
            return True

        if self.start_datetime > other.start_datetime:
            return False

        return self.org_id < other.org_id


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        if self.device_tag is None or self.org_id is None:
            return False

        return CognitoDeviceCredentials.is_valid_tag(self.device_tag) and \
            self.is_valid_deployment_label(self.deployment_label) and self.is_valid_path(self.device_path) and \
            self.is_valid_path(self.location_path)


    def has_unix_era_start(self):
        return self.start_datetime == LocalizedDatetime.unix_era_start()


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict[self.DEVICE_TAG] = self.device_tag
        jdict[self.ORG_ID] = self.org_id
        jdict[self.DEPLOYMENT_LABEL] = self.deployment_label
        jdict[self.DEVICE_PATH] = self.device_path
        jdict[self.LOCATION_PATH] = self.location_path

        jdict[self.START_DATETIME] = None if self.start_datetime is None else self.start_datetime.as_iso8601()
        jdict[self.END_DATETIME] = None if self.end_datetime is None else self.end_datetime.as_iso8601()

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def control_path(self):
        return self.device_path + 'control'


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def username(self):
        return self.device_tag


    @property
    def device_tag(self):
        return self._device_tag


    @property
    def org_id(self):
        return self._org_id


    @property
    def device_path(self):
        return self.__device_path


    @property
    def location_path(self):
        return self.__location_path


    @property
    def start_datetime(self):
        return self._start_datetime


    @property
    def end_datetime(self):
        return self.__end_datetime


    @end_datetime.setter
    def end_datetime(self, end_datetime):
        self.__end_datetime = end_datetime

    @start_datetime.setter
    def start_datetime(self, start_datetime):
        self._start_datetime = start_datetime


    @property
    def deployment_label(self):
        return self.__deployment_label


    @deployment_label.setter
    def deployment_label(self, deployment_label):
        self.__deployment_label = deployment_label


    # ------------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "OrganisationDevice:{device_tag:%s, org_id:%s, device_path:%s, location_path:%s, " \
               "start_datetime:%s, end_datetime:%s, deployment_label:%s}" % \
            (self.device_tag, self.org_id, self.device_path, self.location_path,
             self.start_datetime, self.end_datetime, self.deployment_label)



"""
Created on 2 Dec 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example request:
{"sample": {"tag": "test", "rec": "2020-11-11T16:16:34Z",
"val": {"NO2": {"weV": 0.31569, "aeV": 0.3095, "weC": 0.00395, "cnc": 11.6},
"CO": {"weV": 0.34176, "aeV": 0.25519, "weC": 0.09502, "cnc": 397.6},
"SO2": {"weV": 0.26657, "aeV": 0.26494, "weC": -0.00267, "cnc": 20.5},
"H2S": {"weV": 0.18319, "aeV": 0.26013, "weC": -0.04456, "cnc": 36.7},
"sht": {"hmd": 66.1, "tmp": 22.2}}},
"t-slope": 0.0, "rh-slope": 0.1,
"brd-tmp": 30.3}
"""

import json

from scs_core.comms.uds_client import UDSClient

from scs_core.data.json import JSONify
from scs_core.data.linear_regression import LinearRegression

from scs_core.model.gas.gas_inference_client import GasInferenceClient
from scs_core.model.gas.vB.gas_request import GasRequest

from scs_core.sync.schedule import ScheduleItem


# --------------------------------------------------------------------------------------------------------------------

class VBGasInferenceClient(GasInferenceClient):
    """
    classdocs
    """

    @classmethod
    def construct(cls, socket, inference_uds_path, gas_schedule_item: ScheduleItem):
        # UDS...
        uds_client = UDSClient(socket, inference_uds_path)

        # T / rH slope...
        slope_tally = GasRequest.slope_tally(gas_schedule_item.duration())

        t_regression = LinearRegression(tally=slope_tally)
        rh_regression = LinearRegression(tally=slope_tally)

        return cls(uds_client, t_regression, rh_regression)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, uds_client, t_regression, rh_regression):
        """
        Constructor
        """
        super().__init__(uds_client)

        self.__t_regression = t_regression                  # LinearRegression
        self.__rh_regression = rh_regression                # LinearRegression


    # ----------------------------------------------------------------------------------------------------------------

    def infer(self, gas_sample, board_temp):
        # T / rH slope...
        self.__t_regression.append(gas_sample.rec, gas_sample.sht_datum.temp)
        t_slope = self.__t_regression.slope(default=0.0)

        self.__rh_regression.append(gas_sample.rec, gas_sample.sht_datum.humid)
        rh_slope = self.__rh_regression.slope(default=0.0)

        # request...
        gas_request = GasRequest(gas_sample, t_slope, rh_slope, board_temp)

        # infer...
        self._uds_client.request(JSONify.dumps(gas_request))
        response = self._uds_client.wait_for_response()

        # report...
        return json.loads(response)


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "VBGasInferenceClient:{uds_client:%s, t_regression:%s, rh_regression:%s}" %  \
               (self._uds_client, self.__t_regression, self.__rh_regression)



"""
Created on 4 Mar 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

AQCSV: Parameter codes and associated standard units

NB: initialisation is performed at the foot of this class

example:
{"code": 88374, "description": "OC1 CSN_Rev Unadjusted PM2.5 LC", "unit-code": 105}

https://www.airnow.gov/
"""

import os

from collections import OrderedDict

from scs_core.aqcsv.specification.unit import Unit

from scs_core.csv.csv_archive import CSVArchive
from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class Parameter(CSVArchive, JSONable):
    """
    classdocs
    """

    _retrieved = {}

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def archive_location(cls):
        return os.path.join(os.path.dirname(os.path.realpath(__file__)), 'archive', 'parameters.csv')


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        code = jdict.get('code')
        description = jdict.get('description')
        unit_code = jdict.get('unit-code')

        return Parameter(code, description, unit_code)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, code, description, unit_code):
        """
        Constructor
        """
        self.__code = int(code)                             # int(5)
        self.__description = description                    # string
        self.__unit_code = int(unit_code)                   # int(3)


    def __eq__(self, other):
        try:
            return self.code == other.code and self.description == other.description and \
                   self.unit_code == other.unit_code

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['code'] = self.code
        jdict['description'] = self.description
        jdict['unit-code'] = self.unit_code

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def pk(self):
        return self.code


    def unit(self):
        return Unit.instance(self.unit_code)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def code(self):
        return self.__code


    @property
    def description(self):
        return self.__description


    @property
    def unit_code(self):
        return self.__unit_code


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Parameter:{code:%05d, description:%s, unit_code:%03d}" % \
               (self.code, self.description, self.unit_code)


# --------------------------------------------------------------------------------------------------------------------
# initialisation...

Parameter.retrieve()



"""
Created on 13 Sep 2023

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example document:
{"organisation": {"OrgID": 11, "Label": "CEPEMAR", "LongName": "CEPEMAR", "URL": "https://cepemar.com",
"Owner": "felipe.tatagiba@cepemar.com", "ParentID": null}, "children": [
{"OrgID": 82, "Label": "Vale", "LongName": "Vale S.A.", "URL": "https://www.vale.com/",
"Owner": "felipe.tatagiba@cepemar.com", "ParentID": 11}]}
"""

from collections import OrderedDict

from scs_core.data.array_dict import ArrayDict
from scs_core.data.json import JSONable
from scs_core.data.str import Str


# --------------------------------------------------------------------------------------------------------------------

class OrganisationMembership(JSONable):
    """
    classdocs
    """

    @classmethod
    def merge(cls, organisations, org_children):
        org_dict = ArrayDict([(org_member.parent_id, org_member) for org_member in sorted(org_children)])

        # Accounts...
        return [cls(organisation, org_dict.get(organisation.org_id)) for organisation in organisations]


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, organisation, children):
        """
        Constructor
        """
        self.__organisation = organisation              # Organisation
        self.__children = children                      # array of Organisation


    def __lt__(self, other):
        return self.organisation < other.organisation


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['organisation'] = self.organisation
        jdict['children'] = self.children

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def organisation(self):
        return self.__organisation


    @property
    def children(self):
        return self.__children


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "OrganisationMembership:{organisation:%s, children:%s}" % \
               (self.organisation, Str.collection(self.children))



"""
Created on 3 Sep 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

found from the OPC firmware

example JSON:
{"serial": "177336702", "firmware": "OPC-N3 Iss1.1 FirmwareVer=1.17a...........................BS"}
"""

from collections import OrderedDict

from scs_core.data.json import MultiPersistentJSONable


# --------------------------------------------------------------------------------------------------------------------

class OPCVersion(MultiPersistentJSONable):
    """
    classdocs
    """

    __FILENAME = "opc_version.json"

    @classmethod
    def persistence_location(cls, name):
        filename = cls.__FILENAME if name is None else '_'.join((name, cls.__FILENAME))

        return cls.conf_dir(), filename


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, name=None, skeleton=False):
        if not jdict:
            return None

        serial_no = jdict.get('serial')
        firmware = jdict.get('firmware')

        return cls(serial_no, firmware, name=name)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, serial_no, firmware, name=None):
        """
        Constructor
        """
        super().__init__(name)

        self.__serial_no = serial_no                        # string
        self.__firmware = firmware                          # string


    def __eq__(self, other):                                # ignore name
        try:
            return self.serial_no == other.serial_no and self.firmware == other.firmware

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def serial_no(self):
        return self.__serial_no


    @property
    def firmware(self):
        return self.__firmware


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['serial'] = self.serial_no
        jdict['firmware'] = self.firmware

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "OPCVersion:{name:%s, serial_no:%s, firmware:%s}" %  \
               (self.name, self.serial_no, self.firmware)



"""
Created on 1 Mar 2024

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

aws lambda add-permission --function-name "arn:aws:lambda:us-west-2:696437392763:function:CognitoLogin:Production" \
--source-arn "arn:aws:execute-api:us-west-2:696437392763:lnh2y9ip75/*/*/CognitoLogin/user" \
--principal apigateway.amazonaws.com --statement-id 8ad4f802-a2c0-4e30-948c-416477a6f8c4 \
--action lambda:InvokeFunction

https://docs.aws.amazon.com/apigateway/latest/developerguide/how-to-custom-domains.html
https://docs.aws.amazon.com/apigateway/latest/developerguide/how-to-custom-domains-prerequisites.html

https://docs.aws.amazon.com/apigateway/latest/developerguide/stage-variables.html
https://docs.aws.amazon.com/apigateway/latest/developerguide/how-to-set-stage-variables-aws-console.html
https://docs.aws.amazon.com/apigateway/latest/developerguide/aws-api-gateway-stage-variables-reference.html
Route 53 is unnecessary

EndpointAccess should be initialised in scs_host.sys.__init__

example document:
{"use-default-urls": false, "report-urls": true}
"""

from abc import ABC, abstractmethod
from collections import OrderedDict

from scs_core.data.json import PersistentJSONable
from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class EndpointAccess(PersistentJSONable):
    """
    classdocs
    """

    __FILENAME = "endpoint_access.json"

    @classmethod
    def persistence_location(cls):
        return cls.aws_dir(), cls.__FILENAME


    # ----------------------------------------------------------------------------------------------------------------

    __CONFIG = None

    @classmethod
    def init(cls, manager):
        cls.__CONFIG = cls.load(manager, skeleton=True)


    @classmethod
    def config(cls):
        return cls.default() if cls.__CONFIG is None else cls.__CONFIG


    @classmethod
    def default(cls):
        return cls(False, False)


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return cls.default() if skeleton else None

        use_default_urls = jdict.get('use-default-urls', False)
        report_urls = jdict.get('report-urls', False)

        return cls(use_default_urls, report_urls)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, use_default_urls, report_urls):
        """
        Constructor
        """
        super().__init__()

        self.__use_default_urls = bool(use_default_urls)            # bool
        self.__report_urls = bool(report_urls)                      # bool


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['use-default-urls'] = self.use_default_urls
        jdict['report-urls'] = self.report_urls

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def use_default_urls(self):
        return self.__use_default_urls


    @property
    def report_urls(self):
        return self.__report_urls


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "EndpointAccess:{use_default_urls:%s, report_urls:%s}" %  (self.use_default_urls, self.report_urls)


# --------------------------------------------------------------------------------------------------------------------

class Endpoint(ABC):
    """
    classdocs
    """

    DEFAULT_AUTH = '@southcoastscience.com'

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def header(cls, accept='text/json', auth=None, token=None):
        header = {"Content-type": "application/x-www-form-urlencoded", "Accept": accept}

        if auth is not None:
            header["Authorization"] = auth

        if token is not None:
            header["Token"] = token

        return header


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    @abstractmethod
    def configuration(cls):
        pass


    @classmethod
    def url(cls, *path_extensions):
        return cls.configuration().selected_url(*path_extensions)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, production_path, raw_url):
        self.__production_path = production_path                # string
        self.__raw_url = raw_url                                # string

        self.__access = EndpointAccess.config()


    # ----------------------------------------------------------------------------------------------------------------

    def selected_url(self, *path_extensions):
        url = self.raw_url if self.access.use_default_urls else self.production_url
        extended_url = '/'.join([url] + [str(extension) for extension in path_extensions])

        if self.access.report_urls:
            Logging.getLogger().info('API: %s' % extended_url)

        return extended_url


    @property
    @abstractmethod
    def production_url(self):
        pass


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def production_path(self):
        return self.__production_path


    @property
    def raw_url(self):
        return self.__raw_url


    @property
    def access(self):
        return self.__access


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return (self.__class__.__name__ + ":{production_path:%s, raw_url:%s, access:%s}" %
                (self.production_path, self.raw_url, self.access))


# --------------------------------------------------------------------------------------------------------------------

class APIEndpoint(Endpoint, ABC):
    """
    classdocs
    """

    API_DOMAIN_NAME =   'api.southcoastscience.com'

    # ----------------------------------------------------------------------------------------------------------------

    @property
    def production_url(self):
        return 'https://' + self.API_DOMAIN_NAME + '/' + self.production_path


# --------------------------------------------------------------------------------------------------------------------

class AWSEndpoint(Endpoint, ABC):
    """
    classdocs
    """

    API_DOMAIN_NAME =   'aws.southcoastscience.com'

    # ----------------------------------------------------------------------------------------------------------------

    @property
    def production_url(self):
        return 'https://' + self.API_DOMAIN_NAME + '/' + self.production_path



"""
Created on 5 Jan 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

Getting distance between two points based on latitude/longitude
https://stackoverflow.com/questions/19412462/getting-distance-between-two-points-based-on-latitude-longitude

https://en.wikipedia.org/wiki/Geographic_coordinate_system
"""

from math import sin, cos, sqrt, atan2, radians
from numbers import Number

from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class Position(JSONable):
    """
    classdocs
    """

    R = 6373.0              # approximate radius of earth in Km

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        lat = jdict[0]
        lng = jdict[1]

        if lat == '':
            lat = None

        if lng == '':
            lng = None

        return Position(lat, lng)


    @classmethod
    def construct_from_gga(cls, gga):
        if gga is None:
            return None

        loc = gga.loc

        lat = None if loc is None else loc.deg_lat()
        lng = None if loc is None else loc.deg_lng()

        return Position(lat, lng)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, lat, lng):
        """
        Constructor
        """
        self.__lat = lat                    # float     degrees north of the equator (±90)
        self.__lng = lng                    # float     degrees east of the Greenwich meridian (±180)


    # ----------------------------------------------------------------------------------------------------------------
    # Support for averaging...

    def __add__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError(other)

        lat = self.lat + other.lat
        lng = self.lng + other.lng

        return Position(lat, lng)


    def __truediv__(self, other):
        if not isinstance(other, Number):
            raise TypeError(other)

        lat = self.lat / other
        lng = self.lng / other

        return Position(lat, lng)


    # ----------------------------------------------------------------------------------------------------------------

    def distance(self, other):
        if other is None:
            return None

        if not isinstance(other, self.__class__):
            raise TypeError(other)

        if other.lat is None or other.lng is None:
            return None

        lat1 = radians(self.lat)
        lng1 = radians(self.lng)

        lat2 = radians(other.lat)
        lng2 = radians(other.lng)

        delta_lat = lat2 - lat1
        delta_lng = lng2 - lng1

        a = sin(delta_lat / 2) ** 2 + cos(lat1) * cos(lat2) * sin(delta_lng / 2) ** 2
        c = 2 * atan2(sqrt(a), sqrt(1 - a))
        distance = self.R * c

        return round(distance, 3)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jlist = (
            None if self.lat is None else round(self.lat, 8),
            None if self.lng is None else round(self.lng, 8)
        )

        return jlist


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def lat(self):
        return self.__lat


    @property
    def lng(self):
        return self.__lng


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Position:{lat:%s, lng:%s}" % (self.lat, self.lng)



"""
Created on 22 May 2023

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

document example:
{"device_tag": "scs-ap1-303"}
"""

import requests

from scs_core.aws.client.api_client import APIClient
from scs_core.aws.config.endpoint import APIEndpoint

from scs_core.data.json import JSONable, JSONify


# --------------------------------------------------------------------------------------------------------------------

class Endpoint(APIEndpoint):
    @classmethod
    def configuration(cls):
        return cls('DevSecAPI/DeviceWhitelist',
                   'https://i8lwsjuksc.execute-api.us-west-2.amazonaws.com/default/DeviceWhitelist')


# --------------------------------------------------------------------------------------------------------------------

class DeviceWhitelistManager(APIClient):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self):
        super().__init__()


    # ----------------------------------------------------------------------------------------------------------------

    def find_all(self, token):
        response = requests.get(Endpoint.url(), headers=self._token_headers(token))
        self._check_response(response)

        return sorted([DeviceWhitelistItem.construct_from_jdict(jdict) for jdict in response.json()])


    def create(self, token, device_tag):
        payload = DeviceWhitelistItem(device_tag)

        response = requests.post(Endpoint.url(), headers=self._token_headers(token), data=JSONify.dumps(payload))
        self._check_response(response)


    def delete(self, token, device_tag):
        payload = DeviceWhitelistItem(device_tag)

        response = requests.delete(Endpoint.url(), headers=self._token_headers(token), data=JSONify.dumps(payload))
        self._check_response(response)


# --------------------------------------------------------------------------------------------------------------------

class DeviceWhitelistItem(JSONable):
    """
    classdocs
    """

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        device_tag = jdict.get('device-tag')

        return cls(device_tag)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, device_tag):
        """
        Constructor
        """
        self.__device_tag = device_tag                              # string


    def __lt__(self, other):
        return self.device_tag < other.device_tag


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        return {'device-tag': self.device_tag}


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def device_tag(self):
        return self.__device_tag


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "DeviceWhitelistItem:{device_tag:%s}" % self.device_tag



"""
Created on 24 Apr 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

from scs_core.gas.pid.pid import PID
from scs_core.gas.pid.pid_temp_comp import PIDTempComp


# --------------------------------------------------------------------------------------------------------------------

PIDTempComp.init()      # must be initialised before sensors
PID.init()



"""
Created on 7 Jul 2023

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://stackoverflow.com/questions/36414920/how-to-store-time-without-date-in-python-so-its-convenient-to-compare
https://en.wikipedia.org/wiki/Cron

document example:
{"type": "diurnal", "start": "09:00:00", "end": "17:00:00", "timezone": "Europe/London"}
"""

import pytz
import re

from collections import OrderedDict
from datetime import datetime, time

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONable
from scs_core.data.period import Period
from scs_core.data.timedelta import Timedelta


# --------------------------------------------------------------------------------------------------------------------

class DiurnalPeriod(Period, JSONable):
    """
    classdocs
    """

    __TYPE = 'diurnal'

    @classmethod
    def type(cls):
        return cls.__TYPE


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def is_valid_time(cls, time_str):
        try:
            cls.__time(time_str)
            return True

        except ValueError:
            return False


    @classmethod
    def __time(cls, time_str):
        try:
            return time.fromisoformat(cls.__iso_time_str(time_str))
        except (TypeError, ValueError):
            raise ValueError(time_str)


    @classmethod
    def __iso_time_str(cls, time_str):
        if re.match(r'\d{2}:\d{2}:\d{2}', time_str):
            return time_str

        if re.match(r'\d{2}:\d{2}', time_str):
            return time_str + ':00'

        if re.match(r'\d:\d{2}', time_str):
            return '0' + time_str + ':00'

        raise ValueError(time_str)


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        return cls.construct(jdict.get('start'), jdict.get('end'), jdict.get('timezone'))


    @classmethod
    def construct(cls, start_time_str, end_time_str, timezone_str):
        start_time = cls.__time(start_time_str)
        end_time = cls.__time(end_time_str)

        if timezone_str not in pytz.all_timezones:
            raise ValueError(timezone_str)

        timezone = pytz.timezone(timezone_str)

        return cls(start_time, end_time, timezone)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, start_time: time, end_time: time, timezone: pytz.timezone):
        """
        Constructor
        """
        Period.__init__(self, timezone)

        self.__start_time = start_time                  # time
        self.__end_time = end_time                      # time


    def __lt__(self, other):
        if self.start_time < other.start_time:
            return True

        if self.start_time > other.start_time:
            return False

        if self.end_time < other.end_time:
            return True

        if self.end_time > other.end_time:
            return False

        return str(self.timezone) < str(other.timezone)


    def __contains__(self, point: LocalizedDatetime):
        localized_datetime = point.localize(self.timezone)
        localized_time = localized_datetime.datetime.time()

        return self.start_time <= localized_time < self.end_time


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        return self.start_time != self.end_time


    def crosses_midnight(self):
        return self.start_time > self.end_time


    def has_expiring_dst(self):
        now = LocalizedDatetime.now().localize(self.timezone)
        tomorrow = (now + Timedelta(days=1)).localize(self.timezone)

        return now.dst() != tomorrow.dst()


    # ----------------------------------------------------------------------------------------------------------------

    def checkpoint(self):
        return str(self.end_datetime(LocalizedDatetime.now()).datetime.time())


    def cron(self, minutes_offset):
        return '%d %d * * *' % self.__cron_units(minutes_offset)


    def aws_cron(self, minutes_offset):
        return 'cron(%d %d * * ? *)' % self.__cron_units(minutes_offset)


    def __cron_units(self, minutes_offset):
        end_datetime = self.end_datetime(LocalizedDatetime.now())
        offset_end_datetime = (end_datetime + Timedelta(minutes=minutes_offset)).utc()
        dt = offset_end_datetime.datetime

        return dt.minute, dt.hour


    # ----------------------------------------------------------------------------------------------------------------

    def start_datetime(self, point: LocalizedDatetime):
        p_dt = point.datetime
        s_t = self.start_time

        start_dt = datetime(p_dt.year, month=p_dt.month, day=p_dt.day, hour=s_t.hour, minute=s_t.minute)
        start = LocalizedDatetime(self.timezone.localize(start_dt))

        if self.crosses_midnight():
            start -= Timedelta(days=1)

        return start


    def end_datetime(self, point: LocalizedDatetime):
        p_dt = point.datetime
        e_t = self.end_time

        end_dt = datetime(p_dt.year, month=p_dt.month, day=p_dt.day, hour=e_t.hour, minute=e_t.minute)
        end = LocalizedDatetime(self.timezone.localize(end_dt))

        return end


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self, *args, **kwargs):
        jdict = OrderedDict()

        jdict['type'] = self.type()

        jdict['start'] = str(self.start_time)
        jdict['end'] = str(self.end_time)
        jdict['timezone'] = str(self.timezone)

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def start_time(self):
        return self.__start_time


    @property
    def end_time(self):
        return self.__end_time


    # ----------------------------------------------------------------------------------------------------------------

    def __repr__(self):
        return ' '.join((str(self.start_time), str(self.end_time), str(self.timezone)))


    def __str__(self, *args, **kwargs):
        return "DiurnalPeriod:{start_time:%s, end_time:%s, timezone:%s}" % \
            (self.start_time, self.end_time, self.timezone)



"""
Created on 8 Mar 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example:
{"tag": "scs-be2-2", "shared-secret": "secret1",
"topic": "south-coast-science-dev/production-test/device/alpha-bb-eng-000002/control"}
"""

from collections import OrderedDict

from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class MQTTPeer(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        tag = jdict.get('tag')
        shared_secret = jdict.get('shared-secret')
        topic = jdict.get('topic')

        return cls(tag, shared_secret, topic)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, tag, shared_secret, topic):
        """
        Constructor
        """
        self.__tag = tag                                        # string
        self.__shared_secret = shared_secret                    # string
        self.__topic = topic                                    # string


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['tag'] = self.tag
        jdict['shared-secret'] = self.shared_secret
        jdict['topic'] = self.topic

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def tag(self):
        return self.__tag


    @property
    def shared_secret(self):
        return self.__shared_secret


    @shared_secret.setter
    def shared_secret(self, shared_secret):
        self.__shared_secret = shared_secret


    @property
    def topic(self):
        return self.__topic


    @topic.setter
    def topic(self, topic):
        self.__topic = topic


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "MQTTPeer:{tag:%s, shared_secret:%s, topic:%s}" % \
               (self.tag, self.shared_secret, self.topic)



"""
Created on 3 Apr 2023

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example document:
"""

from collections import OrderedDict

from scs_core.data.array_dict import ArrayDict
from scs_core.data.json import JSONable
from scs_core.data.str import Str


# --------------------------------------------------------------------------------------------------------------------

class OPRMembership(JSONable):
    """
    classdocs
    """

    @classmethod
    def merge(cls, oprs, oups):
        org_dict = ArrayDict([(oup.opr_id, oup) for oup in sorted(oups)])

        # OrganisationPathRoots...
        return [cls(opr, org_dict.get(opr.opr_id)) for opr in oprs]


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, opr, memberships):
        """
        Constructor
        """
        self.__opr = opr                                    # OrganisationPathRoot
        self.__memberships = memberships                    # array of OrganisationUserPath


    def __lt__(self, other):
        return self.opr < other.opr


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['opr'] = self.opr
        jdict['memberships'] = self.memberships

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def opr(self):
        return self.__opr


    @property
    def memberships(self):
        return self.__memberships


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "OPRMembership:{opr:%s, memberships:%s}" % \
               (self.opr, Str.collection(self.memberships))



"""
Created on 13 Apr 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example JSON:
{"root-path": "/srv/removable_data_storage", "delete-oldest": true, "write-interval": 0,
"retrospection-limit": "2023-07-25T11:36:30Z"}
"""

from collections import OrderedDict

from scs_core.csv.csv_data_log import CSVDataLog

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import PersistentJSONable

from scs_core.sys.filesystem import FilesystemReport


# --------------------------------------------------------------------------------------------------------------------

class CSVLoggerConf(PersistentJSONable):
    """
    classdocs
    """

    __FILENAME = "csv_logger_conf.json"

    @classmethod
    def persistence_location(cls):
        return cls.conf_dir(), cls.__FILENAME


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        root_path = jdict.get('root-path')
        delete_oldest = jdict.get('delete-oldest')
        write_interval = jdict.get('write-interval')
        retrospection_limit = LocalizedDatetime.construct_from_iso8601(jdict.get('retrospection-limit'))

        return CSVLoggerConf(root_path, delete_oldest, write_interval, retrospection_limit)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, root_path, delete_oldest, write_interval, retrospection_limit):
        """
        Constructor
        """
        super().__init__()

        self.__root_path = root_path                                # string
        self.__delete_oldest = bool(delete_oldest)                  # bool
        self.__write_interval = int(write_interval)                 # int
        self.__retrospection_limit = retrospection_limit            # LocalizedDatetime or None


    def __eq__(self, other):
        try:
            return self.root_path == other.root_path and self.delete_oldest == other.delete_oldest and \
                   self.write_interval == other.write_interval and self.retrospection_limit == other.retrospection_limit

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def csv_log(self, topic_subject, tag=None, timeline_start=None):
        return CSVDataLog(self.root_path, topic_subject, tag=tag, timeline_start=timeline_start)


    def filesystem_report(self):
        return FilesystemReport.construct(self.root_path)


    def utc_retrospection_start(self, utc_timeline_start):
        utc_retrospection_limit = None if self.retrospection_limit is None else self.retrospection_limit.utc_datetime

        if utc_retrospection_limit is None or utc_timeline_start is None:
            return utc_timeline_start

        if utc_timeline_start < utc_retrospection_limit:
            return utc_retrospection_limit

        return utc_timeline_start


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['root-path'] = self.root_path
        jdict['delete-oldest'] = self.delete_oldest
        jdict['write-interval'] = self.write_interval
        jdict['retrospection-limit'] = self.retrospection_limit

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def root_path(self):
        return self.__root_path


    @property
    def delete_oldest(self):
        return self.__delete_oldest


    @property
    def write_interval(self):
        return self.__write_interval


    @property
    def retrospection_limit(self):
        return self.__retrospection_limit


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "CSVLoggerConf:{root_path:%s, delete_oldest:%s, write_interval:%s, retrospection_limit:%s}" %  \
               (self.root_path, self.delete_oldest, self.write_interval, self.retrospection_limit)



"""
Created on 23 Nov 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import requests

from scs_core.aws.client.api_client import APIClient
from scs_core.aws.config.endpoint import APIEndpoint
from scs_core.aws.security.cognito_authentication import AuthenticationResult


# --------------------------------------------------------------------------------------------------------------------

class Endpoint(APIEndpoint):
    @classmethod
    def configuration(cls):
        return cls('CogSecAPI/CognitoLogin',
                   'https://lnh2y9ip75.execute-api.us-west-2.amazonaws.com/default/CognitoLogin')


# --------------------------------------------------------------------------------------------------------------------

class CognitoLoginManager(APIClient):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self):
        super().__init__()


    # ----------------------------------------------------------------------------------------------------------------

    def user_login(self, credentials):
        url = Endpoint.url('user')
        headers = self._auth_headers()

        response = requests.post(url, headers=headers, json=credentials.as_json())
        self._check_response(response)

        return AuthenticationResult.construct_from_res(response)


    def device_login(self, credentials):
        url = Endpoint.url('device')
        headers = self._auth_headers()

        response = requests.post(url, headers=headers, json=credentials.as_json())
        self._check_response(response)

        return AuthenticationResult.construct_from_res(response)
